from numba import njit, u1,u2, i4, i8, u8, types, literal_unroll, generated_jit
from numba.typed import List
from numba.types import ListType, unicode_type
from cre.caching import unique_hash_v, source_to_cache, import_from_cached, source_in_cache, get_cache_path
from cre.fact import base_fact_field_dict, BaseFact, FactProxy, Fact, lines_in_type_registry, add_to_type_registry, add_type_pickle
from cre.fact_intrinsics import fact_lower_setattr, _register_fact_structref
from cre.obj import cre_obj_get_item, CREObjType, CREObjProxy, CREObjTypeClass, member_info_type
from cre.utils import _struct_get_attr_offset, _sizeof_type, _struct_get_data_ptr, _load_ptr, _struct_get_attr_offset, _obj_cast_codegen, encode_idrec, decode_idrec, _incref_structref, _get_member_offset
# from cre.primitive import Primitive
from cre.context import cre_context
from numba.core.imputils import (lower_cast, numba_typeref_ctor)
from numba.experimental.structref import new, define_boxing, define_attributes, StructRefProxy
from numba.core.extending import overload, intrinsic, lower_builtin, type_callable
from numba.core.typing.typeof import typeof
from cre.core import T_ID_TUPLE_FACT, register_global_default
import cloudpickle

##### NOTE NOTE NTOE TODO -- Need distribute "members" across in order to getitem

# tup_fact_field_dict = {
#     **base_fact_field_dict,
#     # "members": types.Any,
#     # "chr_mbrs_infos" : types.UniTuple(member_info_type,1),
# }

# tup_fact_fields = [(k,v) for k,v in tup_fact_field_dict.items()]

def _up_cast_helper(x):
    if(isinstance(x, CREObjTypeClass)):
        return CREObjType
    else:
        return x

def tf_field_dict_from_types(member_types):   
    member_types = member_types.types if isinstance(member_types, types.Tuple) else member_types
    member_types = tuple([x for x in member_types])
    member_info_tup_type = types.UniTuple(member_info_type,len(member_types))
    # member_t_ids = tuple([_resolve_t_id_helper(x) for x in member_types])
    member_field_dict = {f"a{i}":m for i,m in enumerate(member_types)} 
    # tf_d = {**tup_fact_field_dict,"members" : types.Tuple(member_types),"chr_mbrs_infos":member_info_tup_type}
    tf_d = {**base_fact_field_dict,**member_field_dict}#,**{"chr_mbrs_infos":member_info_tup_type}}
    # print("tf_d", tf_d)
    return tf_d

def tf_get_item(tf,typ, index):
    return cre_obj_get_item(tf.asa(CREObjType),typ,index)

from cre.fact import _gen_props, _gen_getter_jit, _gen_setter_jit
def gen_tuple_fact_source(member_types, TF_T_ID, specialization_name=None, ind='    '):    
    # attr_offsets = get_offsets_from_member_types(member_types)
    base_fields = [(k,v) for k,v in base_fact_field_dict.items()]
    getter_jits = "\n".join([_gen_getter_jit("SpecializedTF",t,attr) for attr,t in base_fields])
    setter_jits = "\n".join([_gen_setter_jit("SpecializedTF",attr,a_id) for a_id, (attr,t) in enumerate(base_fields)])
    properties = "\n".join([_gen_props(attr) for attr,t in base_fields])
    init_fields = f'\n{ind}'.join([f"fact_lower_setattr(st, 'a{i}', members[{i}])" for i in range(len(member_types))])
    # print("::", member_types)
# attr_offsets = np.array({attr_offsets!r},dtype=np.int16)    
# SpecializedTF({"types.StarArgTuple(member_types)" if len(member_types) > 0 else ""})
    return (
f'''from numba import njit, u1, u8, types
from numba.types import UniTuple
from numba.experimental.structref import new
from numba.extending import overload, lower_cast
from cre.fact import uint_to_inheritance_bytes
from cre.fact_intrinsics import define_boxing, get_fact_attr_ptr, _register_fact_structref, fact_mutability_protected_setattr, fact_lower_setattr, _fact_get_chr_mbrs_infos
from cre.tuple_fact import TupleFactClass, TupleFactProxy, tf_field_dict_from_types, tf_get_item, T_ID_TUPLE_FACT{", TupleFact" if specialization_name else ""}
from cre.utils import cast, encode_idrec, _get_member_offset
from cre.obj import member_info_type, set_chr_mbrs
import cloudpickle
TF_T_ID = T_ID_TUPLE_FACT#{TF_T_ID}
member_types = cloudpickle.loads({cloudpickle.dumps(member_types)})
n_members = len(member_types)
field_list = [(k,v) for k,v in tf_field_dict_from_types(member_types).items()]

inheritance_bytes = tuple(list(uint_to_inheritance_bytes(T_ID_TUPLE_FACT)) + [u1(0)] + list(uint_to_inheritance_bytes({TF_T_ID}))) 
num_inh_bytes = len(inheritance_bytes)

@_register_fact_structref
class SpecializedTFClass(TupleFactClass):
    t_id = TF_T_ID
    def __str__(self):
        return '{specialization_name if specialization_name else "TupleFact"}'

    __repr__ = __str__

SpecializedTF = fact_type = SpecializedTFClass(field_list)
SpecializedTF_w_mbr_infos = SpecializedTFClass(field_list+
[("chr_mbrs_infos", UniTuple(member_info_type,{len(member_types)})),
 ("num_inh_bytes", u1),
 ("inh_bytes", UniTuple(u1, num_inh_bytes))])



SpecializedTF._fact_name = "TupleFact"
SpecializedTF.t_id = TF_T_ID
# SpecializedTF._attr_offsets = attr_offsets
SpecializedTF._fact_type_class = SpecializedTFClass
{f'SpecializedTF._specialization_name = "{specialization_name}"' if(specialization_name is not None) else ''}

{(f"""SpecializedTF.parent_type = TupleFact
@lower_cast(SpecializedTF, TupleFact)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)                        
""") if specialization_name is not None else ""
}

default_idrec  = encode_idrec(TF_T_ID, 0, 0xFF)
@njit(cache=True)
def ctor({"*members" if len(member_types) > 0 else ""}):
    st = new(SpecializedTF_w_mbr_infos)
    fact_lower_setattr(st, 'idrec', default_idrec)
    fact_lower_setattr(st, 'hash_val', 0)
    set_chr_mbrs(st, ({", ".join([f"'a{i}'" for i in range(len(member_types))])}))
    fact_lower_setattr(st,'num_inh_bytes', num_inh_bytes)
    fact_lower_setattr(st,'inh_bytes', inheritance_bytes)
    {init_fields}
    return cast(st, SpecializedTF)

SpecializedTFClass._ctor = (ctor,)

{getter_jits}

{setter_jits}

class SpecializedTFProxy(TupleFactProxy):
    __numba_ctor = ctor
    _fact_type = SpecializedTF
    _fact_type_class = SpecializedTFClass
    t_id = TF_T_ID

    def __repr__(self):
        return f"TupleFact({{', '.join([repr(tf_get_item(self,mt,i)) for i,mt in enumerate(member_types)])}})"

    def __str__(self):
        return f"TF({{', '.join([str(tf_get_item(self,mt,i)) for i,mt in enumerate(member_types)])}})"

    def __getitem__(self, i):
        mt = member_types[i]
        return tf_get_item(self,mt,i)
{properties}

@overload(SpecializedTFProxy, prefer_literal=False)
def overload_tup_fact_ctor(*args):
    def impl(*args):
        return ctor(*args)
    return impl

SpecializedTF._fact_proxy = SpecializedTFProxy
SpecializedTF._proxy_class = SpecializedTFProxy

define_boxing(SpecializedTFClass,SpecializedTFProxy)

''')

# f'TupleFact({', '.join([repr_tf_item(self,member_types[i],i) for i in range()])})'

def define_tuple_fact(member_types, context=None, return_proxy=False, return_type_class=False):   
    from cre.func import CREFuncTypeClass, CREFuncType
    # print("::", member_types)
    
    member_types = list(member_types)
    for i, x in enumerate(member_types):
        # Don't specialize CREFunc types 
        if(isinstance(x, CREFuncTypeClass)):
            member_types[i] = CREFuncType
    # print("::>", member_types)


    if(len(member_types) > 0):
        typ_assigments = ", ".join([str(t) for t in member_types])
        specialization_name = f"TupleFact({typ_assigments})"
    else:
        specialization_name = None

    TF_hash_code = unique_hash_v(["TupleFact",member_types])
    if(not source_in_cache("TupleFact",TF_hash_code)):
        # Possible for other types to be defined while running the Fact source
        #  so preregister the t_id then add the pickle later.
        TF_T_ID = add_to_type_registry("TupleFact", TF_hash_code)

        # TupleFacts all have the same t_id
        source = gen_tuple_fact_source(member_types, TF_T_ID, specialization_name)
        source_to_cache("TupleFact", TF_hash_code, source)
        tf_type = import_from_cached("TupleFact", TF_hash_code, ["SpecializedTF"])["SpecializedTF"]
        add_type_pickle(tf_type, TF_T_ID)
        
    else:
        tf_type = import_from_cached("TupleFact", TF_hash_code, ["SpecializedTF"])["SpecializedTF"]

    if(specialization_name is not None):
        context = cre_context(context)
        context._register_fact_type(specialization_name, tf_type, inherit_from=TupleFact)


    out = [tf_type]
    if(return_proxy): out.append(tf_type._fact_proxy)
    if(return_type_class): out.append(tf_type._fact_type_class)
    return tuple(out) if len(out) > 1 else out[0]

@_register_fact_structref
class TupleFactClass(Fact):
    _tf_type_cache = {}
    def __new__(cls, *args):
        # If args is a bunch of types then generate the a definition for the
        #   TupleFact with members_types==args. e.g. tf_type = TupleFact(i8,i8)
        if(len(args) == 0 or isinstance(args[0],types.Type)):
            return define_tuple_fact(args)

        # If args contains one list then assume it is the field list 
        #    and return the corresponding specialization
        if(len(args)==1 and isinstance(args[0],(list,tuple))):
            fields = args[0]
            self = super().__new__(cls)
            super().__init__(self,'TupleFact',fields)
            return self

        else:
            # Otherwise make a new TupleFact
            typ = define_tuple_fact(tuple(typeof(x) for x in args))
            inst = typ._ctor[0](*args)
            return inst

    def __init__(self, *args):
        pass

    def preprocess_fields(self, fields):
        return tuple((name, types.unliteral(typ)) for name, typ in fields)



@lower_cast(TupleFactClass, BaseFact)
@lower_cast(TupleFactClass, CREObjType)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty, incref=False)

from numba.experimental.structref import _Utils, imputils
from numba.core import cgutils, utils as numba_utils

@intrinsic
def _tup_fact_get_chr_mbrs_infos(typingctx, tf_type):
    from cre.context import CREContext
    context = CREContext.get_default_context()

    members_type = [v for k,v in tf_type._fields if k == 'members'][0]
    t_ids = [context.get_t_id(_type=x) for x in members_type.types]

    count = members_type.count
    member_infos_out_type = types.UniTuple(member_info_type, count)

    def codegen(context, builder, sig, args):
        [tf,] = args

        utils = _Utils(context, builder, tf_type)

        baseptr = utils.get_data_pointer(tf)
        baseptr_val = builder.ptrtoint(baseptr, cgutils.intp_t)
        dataval = utils.get_data_struct(tf)
        index_of_members = dataval._datamodel.get_field_position("members")

        member_infos = []
        for i in range(count):
            member_ptr = builder.gep(baseptr, [cgutils.int32_t(0), cgutils.int32_t(index_of_members), cgutils.int32_t(i)], inbounds=True)
            member_ptr = builder.ptrtoint(member_ptr, cgutils.intp_t)
            offset = builder.trunc(builder.sub(member_ptr, baseptr_val), cgutils.ir.IntType(16))
            t_id = context.get_constant(u2, t_ids[i])
            member_infos.append(context.make_tuple(builder, member_info_type, (t_id, offset) ))

        ret = context.make_tuple(builder,member_infos_out_type, member_infos)
        return ret

    sig = member_infos_out_type(tf_type)
    return sig, codegen


class TupleFactProxy(FactProxy):
    def __new__(cls, *args):
        if(len(args) == 0 or isinstance(args[0],types.Type)):
            return define_tuple_fact(args)
        else:
            ctor = getattr(cls,"_ctor",define_tuple_fact(tuple(typeof(x) for x in args))._ctor)[0]
            return ctor(*args)

    def __init__(self,*args):
        pass

    def __repr__(self):
        return str(self)


define_boxing(TupleFactClass, TupleFactProxy)

TupleFact = TupleFactClass()

register_global_default("TupleFact", TupleFact)

@overload(TupleFactProxy, prefer_literal=False,)
def overload_tup_fact_ctor(*args):
    ctor = define_tuple_fact(args)._ctor[0]
    def impl(*args):
        return ctor(*args)
    return impl

TF = TupleFactProxy





from numba import types, njit, guvectorize,vectorize,prange, generated_jit, literally, literal_unroll
from numba.experimental import jitclass
from numba import deferred_type, optional
from numba import void,b1,u1,u2,u4,u8,i1,i2,i4,i8,f4,f8,c8,c16
from numba import types
from numba.typed import List, Dict
from numba.core.types import DictType, ListType, unicode_type, float64, UniTuple, Tuple, Array, optional
from numba.cpython.unicode import  _set_code_point
from numba.experimental import structref
from numba.experimental.structref import new
from numba.extending import overload_method, intrinsic, overload, lower_cast, lower_builtin, SentryLiteralArgs
from numba.core.errors import RequireLiteralValue
from cre.core import TYPE_ALIASES, JITSTRUCTS, py_type_map, numba_type_map, numpy_type_map
from numba.core import types, cgutils
from numba.core.errors import TypingError
from numba.experimental.function_type import _get_wrapper_address
from llvmlite import binding as ll

from cre.caching import unique_hash, source_to_cache, import_from_cached, source_in_cache
import numpy as np
import timeit
import itertools
import types as pytypes
import sys
import cloudpickle
import __main__

from cre.context import CREContextDataType, CREContext, ensure_inheritance, cre_context


# from cre.subscriber import BaseSubscriberType
from cre.structref import define_structref, define_boxing, CastFriendlyStructref
from cre.fact import Fact, BaseFact, cast_fact, get_inheritance_t_ids, resolve_deref_data_ptr
from cre.tuple_fact import TF, TupleFact
from cre.fact_intrinsics import fact_lower_setattr
from cre.utils import (cast, CastFriendlyMixin, lower_setattr, _meminfo_from_struct, decode_idrec, encode_idrec, _call_dtor,
  _ptr_from_struct_incref,  _decref_ptr, _decref_structref, _raw_ptr_from_struct_incref,  _obj_cast_codegen, _incref_structref, 
 _store, _load_ptr, DEREF_TYPE_ATTR, DEREF_TYPE_LIST, _ptr_to_data_ptr, _list_base_from_ptr, new_w_del, _store_safe)
from cre.vector import new_vector, VectorType
from cre.caching import import_from_cached, source_in_cache, source_to_cache
from cre.obj import copy_cre_obj, cre_obj_clear_refs, _iter_mbr_infos, OBJECT_MBR_ID

BASE_T_ID_STACK_SIZE = 16
BASE_F_ID_STACK_SIZE = 64
BASE_FACT_SET_SIZE = 64
BASE_CHANGE_QUEUE_SIZE = 2048
   
#### mem Data Definition ####

memset_fields = {
    # Vector<*Vector<*BaseFact>> i.e. 2D vector that holds pointers to facts
    "facts" : VectorType,

    # Context Data object 
    "context_data" : CREContextDataType,

    # Vector<*Vector<i8>> i.e. 2D vector that holds retracted f_ids
    "retracted_f_ids" : VectorType,

    "names_to_idrecs" : DictType(unicode_type,u8),

    # Vector of change idrecs 
    "change_queue" : VectorType,

    # Placeholder types to ensure that facts and fact vecs are refcounted
    "all_facts" : ListType(BaseFact),

    "all_vecs" : ListType(VectorType),
}

@structref.register
class MemSetTypeClass(CastFriendlyStructref):
    def __str__(self):
        return "cre.MemSetType"
    def __repr__(self):
        return self.__str__()

MemSetType = MemSetTypeClass([(k,v) for k,v in memset_fields.items()])






#### MemSet Definition ####

class MemSet(structref.StructRefProxy):
    ''' '''
    def __new__(cls, context=None, auto_clear_refs=False):
        context = cre_context(context)
        self = memset_ctor(context.context_data, auto_clear_refs)
        self._context = context
        self.indexers = {}
        return self

    # def __init__(self, context=None, auto_clear_refs=False):



    @property
    def context(self):
        if(not hasattr(self,'_context')):
            cd = get_context_data(self)
            self._context = cre_context(cd.name)
        return self._context

    @property
    def change_queue(self):
        return get_change_queue(self)
    
    def declare(self,fact,name=None):
        return memset_declare(self,fact,name)

    def retract(self,identifier):
        return memset_retract(self,identifier)

    def _get_indexer(self, attr):
        indexer = self.indexers.get(attr)
        if(indexer is None):
            indexer = self.indexers[attr] = Indexer(attr)
        return indexer

    def get_facts(self,typ=None, no_subtypes=False,**kwargs):
        self.context._ensure_retro_registers()

        # Implement .get_facts(attr=val)
        if(len(kwargs) > 0):
            attr, val = list(kwargs.items())[0]
            indexer = self._get_indexer(attr)
            return indexer.get_facts(self, val)

        # Implement .get_facts(type)
        if(isinstance(typ,str)):
            typ = self.context.name_to_type[typ]
        return get_facts(self, typ, no_subtypes)

    def __iter__(self):
        return iter(self.get_facts())

    def iter_facts(self,typ):
        self.context._ensure_retro_registers()
        if(isinstance(typ,str)):
            typ = self.context.name_to_type[typ]
        return iter_facts(self, typ)

    def get_fact(self,*args, **kwargs):
        # Implement .get_fact(attr=val)
        if(len(kwargs) > 0):
            attr, val = list(kwargs.items())[0]
            indexer = self._get_indexer(attr)
            return indexer.get_fact(self, val)

        # Implement .get_fact(idrec)
        identifier = args[0]
        if(isinstance(identifier, str)):
            return memset_get_fact_by_name(self, identifier)
        else:
            return memset_get_fact_by_idrec(self, identifier)

    def get_ptr(self):
        return get_ptr(self)

    def modify(self, fact, attr, val):
        return memset_modify(self,fact, attr, val)

    def _repr_helper(self,rfunc,ind="    ",sep="\n",pad='\n'):
        strs = []
        for fact in self.get_facts():
            strs.append(rfunc(fact))
        nl = "\n"
        return f'''MemSet(facts=({pad}{ind}{f"{sep}{ind}".join(strs)}{pad})'''

    def __str__(self,**kwargs):
        # from cre.utils import PrintElapse
        # with PrintElapse("__STR__"):
        return self._repr_helper(str,**kwargs)


    def __repr__(self,**kwargs):
        # from cre.utils import PrintElapse
        # with PrintElapse("__REPR__"):
        return self._repr_helper(repr,**kwargs)

    def __copy__(self):
        cpy = memset_copy(self)
        cpy._context = self._context
        cpy.indexers = {}
        return cpy

    copy = __copy__

    def clear_refs(self):
        memset_clear_refs(self)

    # def __hash__(self):
    #     hash_bytes = memset_long_hash(self)

    def long_hash(self, nchars=30):
        return memset_long_hash_str(self, nchars)
    

    def __del__(self):        
        # NOTE: This definitely has bugs
        try:
            pass
            # memset_dtor(self)
        except Exception as e:
            # If the process is ending then global variables can be sporatically None
            #   thus skip any TypeError of this sort.
            if(isinstance(e, ImportError)): return
            if(isinstance(e, TypeError) and "not callable" in str(e)): return
            
            print("An error occured when trying to clean a cre.MemSet object:\n",e)

    def as_dict(self, key_attr='idrec'):
        d = {}
        if(key_attr):
            for fact in self.get_facts():
                fd = fact.as_dict(key_attr)
                d[fd[key_attr]] = fd
        else:
            for i, fact in enumerate(self.get_facts()):
                d[i] = fact.as_dict(key_attr)
        return d


define_boxing(MemSetTypeClass, MemSet)

@njit(types.void(MemSetType),cache=True)
def memset_clear_refs(ms):
    '''Decref out data structures in ms that we explicitly incref'ed '''
    # print("MEMSET DEL!!!", )
    # ms = _struct_from_ptr(MemSetType, _raw_ptr_from_struct(_ms)-48)
    # _incref_structref(ms)
    # print(ms.facts)
    facts = ms.facts#_load_ptr(VectorType, _raw_ptr_from_struct(_ms))
    #Decref all declared facts and their container vectors 
    for i in range(facts.head):
        facts_ptr = facts.data[i]
        if(facts_ptr == 0): continue
        facts_i = cast(facts_ptr, VectorType)
        for j in range(facts_i.head):
            fact_ptr = facts_i.data[j]
            if(fact_ptr == 0): continue
            # print("ptr", j, fact_ptr)
            fact = cast(fact_ptr, BaseFact)
            cre_obj_clear_refs(fact)
            # _iter_mbr_infos(fact)

            # _decref_structref(fact)
            # _incref_structref(fact)
            # _call_dtor(fact)
            # _decref_ptr(fact_ptr)
        # _decref_ptr(facts_ptr)

    #Decref the inner vectors of retracted_f_ids
    # for i in range(ms.retracted_f_ids.head):
    #     ptr = ms.retracted_f_ids.data[i]
    #     _decref_ptr(ptr)
    # print("END")

# HARD CODE WARNING: Pointer misalignment if layout changes inside nrt.c.
SIZEOF_NRT_MEMINFO = 48
@njit(types.void(i8),cache=True)
def memset_data_clear_refs(ms_data_ptr):
    ms = cast(ms_data_ptr-SIZEOF_NRT_MEMINFO, MemSetType)
    memset_clear_refs(ms)


memset_data_clear_refs_addr = _get_wrapper_address(memset_data_clear_refs, types.void(i8))
ll.add_symbol("CRE_MemsetData_ClearRefs", memset_data_clear_refs_addr)

@njit(cache=True)
def expand_mem_set_types(ms, n):
    ''' Expands facts and retracted_f_ids by n.'''
    for i in range(n):
        v = new_vector(BASE_F_ID_STACK_SIZE)    
        ms.all_vecs.append(v)

        v_ptr = cast(v, i8)
        ms.retracted_f_ids.add(v_ptr)

        v = new_vector(BASE_FACT_SET_SIZE)    
        ms.all_vecs.append(v)
        
        v_ptr = cast(v, i8)
        ms.facts.add(v_ptr)

@njit(MemSetType(CREContextDataType,types.boolean), cache=True)
def memset_ctor(context_data, auto_clear_refs=False):
    if(auto_clear_refs):
        st = new_w_del(MemSetType, "CRE_MemsetData_ClearRefs")
    else:
        st = new(MemSetType)
    st.context_data = context_data
    st.facts = new_vector(BASE_T_ID_STACK_SIZE)
    st.retracted_f_ids = new_vector(BASE_T_ID_STACK_SIZE)
    st.names_to_idrecs = Dict.empty(unicode_type,u8)
    st.change_queue = new_vector(BASE_CHANGE_QUEUE_SIZE)
    st.all_facts = List.empty_list(BaseFact)
    st.all_vecs = List.empty_list(VectorType)
    L = max(len(context_data.parent_t_ids)+1,1)
    expand_mem_set_types(st,L)
    return st

@overload(MemSet)
def overload_MemSet(context_data=None, auto_clear_refs=False):
    if(context_data is None):
        raise ValueError("MemSet() must be provided context_data when instantiated in jitted context.")

    def impl(context_data=None, auto_clear_refs=False):
        if(context_data is None):
            # Compile-time error should prevent  
            raise ValueError()            
        else:
            return memset_ctor(context_data, auto_clear_refs)

    return impl

@njit(cache=True)
def get_context_data(self):
    return self.context_data

@njit(cache=True)
def get_change_queue(self):
    return self.change_queue

@njit(cache=True)
def get_ptr(self):
    return cast(self, i8)

@njit(cache=True)
def facts_for_t_id(ms,t_id):
    L = len(ms.facts)
    if(t_id >= L):
        expand_mem_set_types(ms, (t_id+1)-L)
    # print("Declare", t_id, mem_data.facts[t_id])
    return cast(ms.facts[u8(t_id)], VectorType)

@njit(cache=True)
def fact_at_f_id(typ, t_id_facts,f_id):
    ptr = t_id_facts.data[f_id]
    if(ptr != 0):
        return cast(ptr, typ)
    else:
        return None

@njit(cache=True)
def retracted_f_ids_for_t_id(ms,t_id):
    return cast(ms.retracted_f_ids[u8(t_id)], VectorType)

#### Helper Functions ####

@njit(cache=True)
def make_f_id_empty(ms, t_id, f_id):
    '''Removes fact at t_id, f_id freeing up its f_id. f_id is added 
        to the set of rectracted f_ids so it can be recycled.'''
    facts = facts_for_t_id(ms,t_id)
    fact_ptr = facts[f_id]
    if(fact_ptr != 0):
        retracted_f_ids_for_t_id(ms,t_id).add(f_id)
        # _decref_ptr(fact_ptr)
    
    facts[f_id] = 0


@njit(cache=True)
def next_empty_f_id(ms,facts,t_id):
    '''Gets the next dead f_id from retracting facts otherwise returns 
        a fresh one pointing to the end of the meminfo list'''
    f_id_vec = retracted_f_ids_for_t_id(ms,t_id)
    if(f_id_vec.head <= 0):
        return len(facts) # fresh f_id.

    return f_id_vec.pop() # recycled f_id.


@njit(u2(MemSetType, BaseFact),cache=True)
def resolve_t_id(ms, fact):
    '''Gets a t_id from a fact, and ensure that the MemSet can hold
        facts with that t_id. Also ensures that ms.context_data's
        inheritance structures are up to date, so that we can
        do get_facts(Type) and get all subtypes of Type.    
    '''
    cd = ms.context_data

    # Get the fact's t_id
    t_id, _, _ = decode_idrec(fact.idrec)

    # Ensure that inheritance structures are up to date
    if(t_id >= len(cd.child_t_ids) or len(cd.child_t_ids[u8(t_id)])==0):
        inh_t_ids = get_inheritance_t_ids(fact)
        parent_t_id = -1
        for _t_id in inh_t_ids:
            cd.unhandled_retro_registers.append(u2(_t_id))
            ensure_inheritance(cd, _t_id, parent_t_id)
            parent_t_id = i8(_t_id)

        

    # Ensure that the data structures in MemSet are long enough to index t_id.
    L = len(ms.facts)
    if(t_id >= L):
        expand_mem_set_types(ms, 1+L-t_id)

    return t_id

@njit(cache=True)
def name_to_idrec(ms,name):
    names_to_idrecs = ms.names_to_idrecs
    if(name not in names_to_idrecs):
        raise KeyError("Fact not found.")
    return names_to_idrecs[name]



@njit(u8(MemSetType,BaseFact),cache=True)
def declare_fact(ms, fact):
    '''Declares a fact to a MemSet'''
    
    # Acquire a reference to the fact and get it's t_id
    # fact_ptr = i8(_raw_ptr_from_struct_incref(fact)) #.4ms / 10000
    ms.all_facts.append(fact)
    fact_ptr = cast(fact, i8) #.4ms / 10000
    t_id = resolve_t_id(ms, fact)  #.1ms / 10000
    
    # Get the facts Vector for facts with t_id  
    facts = facts_for_t_id(ms, t_id) #negligible

    # Get the next empty f_id. Retracted f_ids are recycled for cache locality.
    f_id = next_empty_f_id(ms, facts, t_id) # .5ms / 10000

    # Assign a new idrec to fact
    idrec = encode_idrec(t_id,f_id,0) #negligable
    fact.idrec = idrec #negligable

    # Make sure facts is big enough 
    while(f_id >= len(facts.data)): # ??ms / 10000
        facts.expand()

    # Put the fact into facts at f_id. Release any old facts.
    # if(facts.data[f_id] != 0): _decref_ptr(facts.data[f_id])
    facts.set_item_safe(f_id, fact_ptr)
    
    ms.change_queue.add(idrec)

    return idrec

@njit(cache=True)
def declare_name(ms,name,idrec):
    ms.names_to_idrecs[name] = idrec

@njit(cache=True)
def declare_fact_name(ms,fact,name):
    idrec = declare_fact(ms,fact)        
    declare_name(ms,name,idrec)
    return idrec

##### retract #####

@njit(cache=True)
def retract_by_idrec(ms,idrec):
    t_id, f_id, _ = decode_idrec(idrec) #negligible

    fact = memset_get_fact(ms, idrec)
    fact_lower_setattr(fact,'idrec',encode_idrec(t_id,0,u1(-1)))

    make_f_id_empty(ms,i8(t_id), i8(f_id)) #3.6ms
    ms.change_queue.add(encode_idrec(t_id, f_id, u1(0xFF)))

    # signal_subscribers_change(mem, idrec) #.8ms

@njit(cache=True)
def retract_by_name(ms,name):
    idrec = name_to_idrec(ms,name)
    retract_by_idrec(ms,idrec)
    del ms.names_to_idrecs[name]


##### modify #####

@njit(cache=True)
def modify_by_idrec(ms,idrec,attr,val):
    raise NotImplemented()
    #lower_setattr(fact,literally(attr),val)
    #TODO signal_subscribers w/ idrec w/ attr_ind

@njit(cache=True)
def modify(ms,fact,attr,val):
    return ms.modify(fact,attr,val)


#### get_fact ####

@njit(BaseFact(MemSetType,u8), cache=True)
def memset_get_fact_by_idrec(self, idrec):
    t_id, f_id, _ =  decode_idrec(idrec)
    # if(t_id != typ_t_id): raise ValueError(err_msg)
    facts = facts_for_t_id(self, t_id) #negligible
    fact_ptr = facts.data[f_id]
    return cast(fact_ptr, BaseFact)

@njit(BaseFact(MemSetType,unicode_type), cache=True)
def memset_get_fact_by_name(self, name):
    idrec = self.names_to_idrecs[name]
    t_id, f_id, _ =  decode_idrec(idrec)
    # if(t_id != typ_t_id): raise ValueError(err_msg)
    facts = facts_for_t_id(self, t_id) #negligible
    fact_ptr = facts.data[f_id]
    return cast(fact_ptr, BaseFact)

@overload_method(MemSetTypeClass, "get_fact")
def memset_get_fact(self, identifier, typ=None):
    if(isinstance(typ,types.Omitted) or typ is None):
        return_typ = BaseFact    
    else:
        return_typ = typ.instance_type
    if(isinstance(identifier, types.Integer)):
        def impl(self, identifier, typ=None):
            return cast(memset_get_fact_by_idrec(self, identifier), return_typ)

    elif(isinstance(identifier, unicode_type)):
        def impl(self, identifier, typ=None):
            return cast(memset_get_fact_by_name(self, identifier),return_typ)
    return impl


#### overloads for declare, rectract, modify ####

@generated_jit(cache=True)
@overload_method(MemSetTypeClass, "declare")
def memset_declare(self, fact, name=None):    
    if(isinstance(fact, Fact)):
        if(not name or isinstance(name, (types.NoneType,types.Omitted))):
            def impl(self, fact, name=None):
                return declare_fact(self,fact)
        else:
            def impl(self, fact, name=None):
                return declare_fact_name(self,fact,name)
    elif(isinstance(fact, types.BaseTuple)):
        def impl(self, fact, name=None):
            return declare_fact(self,TF(*fact))
    else:
        raise TypingError(f"Cannot declare fact of type '{type(fact)}'.")
    
    return impl

@generated_jit(cache=True)
@overload_method(MemSetTypeClass, "retract")
def memset_retract(self, identifier):
    if(identifier in (str,unicode_type)):
        def impl(self, identifier):
            return retract_by_name(self,identifier)
    elif(isinstance(identifier,types.Integer)):
        def impl(self, identifier):
            return retract_by_idrec(self,u8(identifier))
    elif(isinstance(identifier,types.StructRef)):
        def impl(self, identifier):
            return retract_by_idrec(self,identifier.idrec)
    else:
        raise TypingError(f"Cannot retract fact identifier of type '{type(identifier)}'." +
                        "mem.retract() accepts a valid fact idrec, name, or fact instance.")
    return impl

@generated_jit(cache=True)
@overload_method(MemSetTypeClass, "modify")
def memset_modify(self, fact, attr, val):
    if(not isinstance(fact,types.StructRef)): 
        raise TypingError(f"Modify requires a fact instance, got instance of'{type(fact)}'.")

    SentryLiteralArgs(['attr']).for_function(memset_modify).bind(self, fact, attr, val) 
    a_id = u1(list(fact.field_dict.keys()).index(attr._literal_value))

    def impl(self, fact, attr, val):
        fact_lower_setattr(fact, attr, val)
        #TODO signal_subscribers w/ idrec w/ attr_ind
        # signal_subscribers_change(mem, fact.idrec)
        t_id, f_id, _ = decode_idrec(fact.idrec)
        self.change_queue.add(encode_idrec(t_id, f_id, a_id))
    # if(isinstance(attr, types.Literal)):
    # def impl(self, fact, attr, val):
    #     modify_by_fact(self, fact, attr, val)
    #     # fact_lower_setattr(fact, attr, val)

    return impl



@generated_jit(nopython=True, cache=True)
def memset_modify_w_deref_infos(self, fact, deref_infos, val):
    if(not isinstance(fact,types.StructRef)): 
        raise TypingError(f"Modify requires a fact instance, got instance of'{type(fact)}'.")
    # print("<<", deref_infos)
    val_type = val
    def impl(self, fact, deref_infos, val):
        if(len(deref_infos) == 0):
            raise ValueError("Cannot setattr with empty deref_infos.")

        final_deref_info = deref_infos[-1]
        a_id = u1(final_deref_info.a_id)

        field_data_ptr = resolve_deref_data_ptr(fact, deref_infos)
        t_id, f_id, _ = decode_idrec(fact.idrec)

        field_ptr = field_data_ptr#_load_ptr(val_type, field_data_ptr)
    
        # incref new value
        _incref_structref(val)

        # decref old value (must be last in case new value is old value)
        _decref_ptr(field_ptr)
        # print("D")

        if(val is None):
            # Attr case
            if(final_deref_info.type == DEREF_TYPE_ATTR):
                # print("IS NONE")
                _store(i8, field_data_ptr, 0)

            # List case
            else:
                # TODO: some issue with this 
                raise ValueError("List cannot be set to None")
                new_lst = List.empty_list(val_type) 
                _store(i8, field_data_ptr, _ptr_from_struct_incref(new_lst))
        else:
            # print("IS normal", val)
            _store(val_type, field_data_ptr, val)

        # print("Z")

        self.change_queue.add(encode_idrec(t_id, f_id, a_id))

    return impl



fact_iterator_field_dict = {
    "memset" : MemSetType,
    "t_ids": i8[::1],
    "curr_ind": i8 ,
    "curr_t_id_ind": i8,
    "fact_type": types.Any
}
fact_iterator_fields = [(k,v) for k,v, in fact_iterator_field_dict.items()]

# FactIterator, FactIteratorType = define_structref("FactIterator",fact_iter_fields)

def gen_fact_iter_source(fact_type):
    return f'''import cloudpickle
from numba import njit, u2, types
from numba.experimental.structref import new
from cre.memset import FactIterator, FactIteratorType, fact_iterator_field_dict, fact_iter_next_raw_ptr, GenericFactIteratorType
from cre.utils import cast
fact_type = cloudpickle.loads({cloudpickle.dumps(fact_type)})
f_iter_type = FactIteratorType([(k,v) for k,v in {{**fact_iterator_field_dict ,"fact_type": types.TypeRef(fact_type)}}.items()])
f_iter_type._fact_type = fact_type


# @njit(fact_type(GenericFactIteratorType),cache=True)
# def fact_iterator_next(it):
#     return cast(fact_iter_next_raw_ptr(it), fact_type)
# '''




class FactIterator(structref.StructRefProxy):
    ''' '''
    f_iter_type_cache = {}
    def __new__(cls, ms, fact_type):
        return iter_facts(ms, fact_type)
        
    def __next__(self):
        return fact_iter_next(self)#self.fact_iterator_next(self)

    def __iter__(self):
        return self


from numba.types import (SimpleIteratorType,)

@structref.register
class FactIteratorType(CastFriendlyMixin, types.StructRef):
    def __str__(self):
        return f"cre.FactIterator[{self._fact_type}]"

    @property
    def iterator_type(self):
        return self

    @property
    def yield_type(self):
        return self._fact_type
    # def preprocess_fields(self, fields):
    #     return tuple((name, types.unliteral(typ)) for name, typ in fields)


define_boxing(FactIteratorType, FactIterator)
GenericFactIteratorType = FactIteratorType(fact_iterator_fields)
GenericFactIteratorType._fact_type = BaseFact

@lower_cast(FactIteratorType, GenericFactIteratorType)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty, incref=False)

@njit(GenericFactIteratorType(MemSetType,i8[::1]),cache=True)
def generic_fact_iterator_ctor(ms, t_ids):
    st = new(GenericFactIteratorType)
    st.memset = ms
    st.t_ids = t_ids
    st.curr_ind = 0
    st.curr_t_id_ind = 0
    return st

# @generated_jit(cache=True)
# def fact_iterator_ctor(mem, typ):

@lower_builtin('getiter', FactIteratorType)
def getiter_fact_iter(context, builder, sig, args):
    return args[0].value

@overload_method(FactIteratorType,'__iter__')
def getiter(self):
    def impl(self):
        return self
    return impl

# @overload(FactIterator)
@generated_jit(cache=True)
@overload_method(MemSetTypeClass,'iter_facts')
def iter_facts(ms, fact_type, no_subtypes=False):
    assert isinstance(fact_type, types.TypeRef)

    fact_type = fact_type.instance_type
    fact_num = fact_type._fact_num

    hash_code = unique_hash([fact_type])
    if(not source_in_cache('FactIterator', hash_code)):
        source = gen_fact_iter_source(fact_type)
        source_to_cache('FactIterator', hash_code, source)
    it_type = import_from_cached('FactIterator', hash_code, ['f_iter_type'])['f_iter_type']
    def impl(ms, fact_type, no_subtypes=False):
        cd = ms.context_data
        if(no_subtypes):
            t_ids = np.array((cd.fact_num_to_t_id[fact_num],),dtype=np.uint16)
        else:
            t_ids = cd.child_t_ids[cd.fact_num_to_t_id[fact_num]]
        _it = generic_fact_iterator_ctor(ms,t_ids)
        return cast(_it, it_type)
    return impl
    # context = mem.context



@njit(i8(GenericFactIteratorType),cache=True)
def fact_iter_next_raw_ptr(it):
    while(True):
        if(it.curr_t_id_ind >= len(it.t_ids)): raise StopIteration()
        t_id = it.t_ids[it.curr_t_id_ind]
        facts_ptr = it.memset.facts[i8(t_id)]
        if(facts_ptr != 0):
            facts = cast(facts_ptr, VectorType)
            if(it.curr_ind < len(facts)):
                ptr = facts[it.curr_ind]
                it.curr_ind +=1
                if(ptr != 0): return ptr
            else:
                it.curr_ind = 0
                it.curr_t_id_ind +=1
        else:
            it.curr_ind = 0
            it.curr_t_id_ind += 1


# @generated_jit(cache=True, nopython=True)
# @overload_method(MemoryTypeTemplate,'get_t_id')
# def get_t_id(mem, fact_type, no_subtypes=False):
#     def impl(mem, fact_type, no_subtypes=False):
#         return mem.context_data.fact_num_to_t_id[fact_num]
#     return impl

# @generated_jit(cache=True)
# @overload_method(MemoryTypeTemplate,'get_subtype_t_ids')
# def get_subtype_t_ids(mem, fact_type, no_subtypes=False):
#     def impl(mem, fact_type, no_subtypes=False):
#         return mem.context_data.child_t_ids[cd.fact_num_to_t_id[fact_num]]
#     return impl


@generated_jit(cache=True)
@overload_method(MemSetTypeClass,'get_facts')
def get_facts(ms, fact_type=None, no_subtypes=False):
    if(isinstance(fact_type, types.TypeRef)):
        _fact_type = fact_type.instance_type
    elif(isinstance(fact_type, types.NoneType)):
        _fact_type = BaseFact
    else:
        raise ValueError("fact_type of get_facts() must be a numba type or None.")
    
    
    get_all = fact_type is None or _fact_type is BaseFact

    fact_t_id = _fact_type.t_id
    def impl(ms, fact_type=None, no_subtypes=False):
        cd = ms.context_data
        if(get_all):
            t_ids = np.arange(len(ms.facts), dtype=np.uint16)
        else:
            # t_id = fact_type.t_id#cd.fact_num_to_t_id[fact_num]
            if(no_subtypes):
                t_ids = np.array((fact_t_id,),dtype=np.uint16)
            else:                
                t_ids = cd.child_t_ids[fact_t_id]
        out = List.empty_list(_fact_type)
        curr_t_id_ind = 0
        curr_ind = 0
        while(True):
            if(curr_t_id_ind >= len(t_ids)): break
            t_id = t_ids[curr_t_id_ind]

            if(t_id < len(ms.facts)):                
                facts_ptr = ms.facts[i8(t_id)]
                # print("facts_ptr", facts_ptr)
                if(facts_ptr != 0):
                    facts = cast(facts_ptr, VectorType)
                    # print("facts", t_id, facts.head, facts.data)
                    if(curr_ind < len(facts)):
                        ptr = facts[curr_ind]
                        curr_ind +=1
                        if(ptr != 0): 
                            out.append(cast(ptr, _fact_type))
                    else:
                        curr_ind = 0
                        curr_t_id_ind +=1
                else:
                    curr_ind = 0
                    curr_t_id_ind += 1
            else:
                curr_ind = 0
                curr_t_id_ind += 1
        return out
    return impl

        

@generated_jit(cache=True)
def fact_iter_next(it):
    # it_type = it.instance_type
    fact_type = it._fact_type
    def impl(it):
        ptr = fact_iter_next_raw_ptr(it)
        return cast(ptr, fact_type)
    return impl


# ----------------------------------------------------------------------
# : Indexer - Allows for .get_facts(attr=val)

base_indexer_fields = {
    # Context Data object 
    "head" : i8,
    "mapping" : types.Any,
    "attr" : types.Any,
    "types_t_ids" : types.Any,
}

@structref.register
class IndexerTypeClass(CastFriendlyStructref):
    pass

class Indexer(structref.StructRefProxy):
    def __new__(cls, attr):
        base_types = get_base_types_with_attr(attr)
        attr_type = base_types[0][1]
        ctor, indexer_type = indexer_ctor_impl(base_types, attr)
        self = ctor()

        indexer_get_facts.compile(ListType(BaseFact)(indexer_type, MemSetType, attr_type))
        indexer_get_fact.compile(BaseFact(indexer_type, MemSetType, attr_type))

        self.get_facts_ep = indexer_get_facts.overloads[(indexer_type, MemSetType, attr_type)].entry_point
        self.get_fact_ep = indexer_get_fact.overloads[(indexer_type, MemSetType, attr_type)].entry_point
        return self

    def get_facts(self, memset, val):
        return self.get_facts_ep(self, memset, val)


    def get_fact(self, memset, val):
        return self.get_fact_ep(self, memset, val)




define_boxing(IndexerTypeClass, Indexer)

def get_base_types_with_attr(attr):
    context = cre_context()
    context.children_of
    base_types = {}

    for fact_type, parents in context.parents_of.items():
        if(isinstance(fact_type, Fact)):
            base_type = fact_type
            if(not hasattr(base_type,'spec')): continue
            while(hasattr(base_type,'parent_type') and
                  hasattr(base_type.parent_type, 'spec') and
                  attr in base_type.parent_type.spec):
                base_type = base_type.parent_type
            if(attr in base_type.spec and base_type not in base_types):
                val_type = base_type.spec[attr]['type']
                # child_t_ids = context.context_data.child_t_ids[base_type.t_id]
                # child_t_ids = types.literal(tuple([t_id for t_id in child_t_ids]))
                base_types[base_type] = val_type

    out = Tuple(tuple([Tuple((bt, vt)) for bt,vt in base_types.items()]))
    return out

vec2_type = Tuple((VectorType,VectorType))

# TODO: Should probably see if there is a way to make more generic 
#  version of Indexer
# def new_indexer(attr):
#     base_types = get_base_types_with_attr(attr)
#     return indexer_ctor_impl(base_types, attr)()

_indexer_ctor_impls = {}
def indexer_ctor_impl(base_types, attr):
    # SentryLiteralArgs(['attr']).for_function(indexer_ctor).bind(base_types, attr)
    # _base_types = base_types.instance_type
    tup = (base_types, attr)
    if(tup not in _indexer_ctor_impls):
        if(len(base_types) == 0):
            context = cre_context()
            raise ValueError(f"No fact_types with attribute {attr!r} defined in context {context.name!r}.")
        val_type = None
        for _,_val_type in base_types:    
            if(val_type is not None and _val_type != val_type):
                raise ValueError("Cannot use indexer on attribute that differs in type across two fact types.")
            val_type = _val_type

        indexer_fields = {
            **base_indexer_fields, 
            "mapping" : DictType(val_type, vec2_type),
            "inv_mapping" : DictType(u8, val_type),
            "attr" : types.literal(attr),
            "base_types" : types.TypeRef(base_types),
        }
        # print(base_types)
        indexer_type = IndexerTypeClass([(k,v) for k,v in indexer_fields.items()])
        # print(indexer_type)        
        indexer_type._base_types = base_types

        long_hash = unique_hash(tup)

        if(not source_in_cache('MemSetIndexer', long_hash)):
            types_bytes = cloudpickle.dumps((indexer_type, val_type))
            source = f'''
import cloudpickle
from numba import u8, types, njit 
from numba.typed import Dict
from numba.experimental.structref import new
from cre.vector import VectorType

vec2_type = types.Tuple((VectorType,VectorType))
indexer_type, val_type = cloudpickle.loads({types_bytes})
@njit(indexer_type(), cache=True)
def ctor():
    st = new(indexer_type)
    st.head = 0
    st.mapping = Dict.empty(val_type, vec2_type)
    st.inv_mapping = Dict.empty(u8, val_type)
    return st
'''
            source_to_cache('MemSetIndexer', long_hash, source)
        ctor = import_from_cached('MemSetIndexer', long_hash, ['ctor'])['ctor']
        # print(ctor)
        _indexer_ctor_impls[tup] = (ctor, indexer_type)
    return _indexer_ctor_impls[tup]



from cre.fact_intrinsics import fact_lower_getattr
@njit(cache=True)
def indexer_update_declare(self, change_event, fact): 
    val = fact_lower_getattr(fact, self.attr)

    if(val not in self.mapping):
        self.mapping[val] = (new_vector(1), new_vector(1))
    idrec_vec, hole_vec = self.mapping[val]

    # Don't re-add on retract-declare pair
    if(not change_event.was_retracted):
        if(len(hole_vec) > 0):
            idrec_vec[hole_vec.pop()] = change_event.idrec
        else:
            idrec_vec.add(change_event.idrec)

    self.inv_mapping[change_event.idrec] = val

@njit(cache=True)
def indexer_update_retract(self, change_event): 
    val = self.inv_mapping[change_event.idrec]
    idrec_vec, hole_vec = self.mapping[val]
    index = np.nonzero(idrec_vec.data==change_event.idrec)[0][0]
    hole_vec.add(index)
    idrec_vec[index] = 0

    del self.inv_mapping[change_event.idrec]


from cre.change_event import accumulate_change_events
@generated_jit(cache=True, nopython=True)
def indexer_update(self, memset):
    context_data = cre_context().context_data
    base_types_t_ids = []
    for bt,vt in self._base_types:
        base_types_t_ids.append((bt,tuple(context_data.child_t_ids[bt.t_id])))
    base_types_t_ids = tuple(base_types_t_ids)
    def impl(self, memset):
        cq = memset.change_queue
        if(self.head == cq.head): return

        for change_event in accumulate_change_events(cq, self.head, cq.head):
            for tup in literal_unroll(base_types_t_ids):
                base_type, t_ids = tup
                if(change_event.t_id in t_ids):
                    if(change_event.was_declared):
                        facts = facts_for_t_id(memset, change_event.t_id)
                        fact = cast(facts[change_event.f_id], base_type)
                        indexer_update_declare(self, change_event, fact)

                    elif(change_event.was_retracted):
                        indexer_update_retract(self, change_event)
                        
        self.head = cq.head
    return impl

@njit(cache=True)
def indexer_get_facts(self, memset, val):
    if(self.head != memset.change_queue.head):
        indexer_update(self, memset)
    facts = List.empty_list(BaseFact)
    if(val in self.mapping):
        idrecs, _ = self.mapping[val]
        for i in range(idrecs.head):
            idrec = u8(idrecs.data[i])
            if(idrec != 0):
                facts.append(memset_get_fact_by_idrec(memset,idrec))
    return facts

@njit(cache=True)
def indexer_get_fact(self, memset, val):
    if(self.head != memset.change_queue.head):
        indexer_update(self, memset)
    if(val in self.mapping):
        idrecs, _ = self.mapping[val]
        for i in range(idrecs.head):
            idrec = u8(idrecs.data[i])
            if(idrec != 0):
                return memset_get_fact_by_idrec(memset,idrec)
    raise KeyError("No fact found.")





# ----------------------------------------------------------------------
# : memset_copy - Allows for .get_facts(attr=val)


@njit(cache=True)
def memset_copy(self):
    new = memset_ctor(self.context_data, False)    

    new_facts = List.empty_list(BaseFact)
    ptr_map = Dict.empty(i8,i8)
    for i in range(self.facts.head):
        facts_ptr = self.facts.data[i]
        if(facts_ptr != 0):
            facts = cast(facts_ptr, VectorType)
            for j in range(facts.head):
                fact_ptr = facts.data[j]
                if(fact_ptr != 0):
                    fact = cast(fact_ptr, BaseFact)
                    new_fact = copy_cre_obj(fact)
                    ptr_map[fact_ptr] = cast(new_fact,i8)
                    new.declare(new_fact)
                    new_facts.append(new_fact)

    # Copy any fact members as weak pointers
    for new_fact in new_facts:
        for t_id, m_id, data_ptr in _iter_mbr_infos(new_fact):
            if(m_id == OBJECT_MBR_ID):
                old_fact_ptr = _load_ptr(i8, data_ptr)
                if(old_fact_ptr != 0 and old_fact_ptr in ptr_map):
                    new_fact = cast(ptr_map[old_fact_ptr],BaseFact)
                    _store_safe(BaseFact, data_ptr, new_fact)

    # TODO should copy any lists of facts
    
    return new


from cre.hashing import accum_item_hash
from numba.cpython.hashing import _Py_hash_t, _PyHASH_XXPRIME_5
from numba import u8
from cre.type_conv import byte_string_to_base64

@njit(cache=True, locals={"i": u8, "j": u8, "c": u8, "nc": u8, "mx": u8,"rest": u8})
def memset_long_hash_bytes(ms, byte_width=24):
    facts = get_facts(ms, None)
    fact_hashes = np.empty(len(facts), dtype=np.uint64)
    for i, fact in enumerate(facts):
        fact_hashes[i] = hash(fact)

    hash_bytes = np.empty(byte_width, dtype=np.uint8)
    buffer0 = np.empty(1, dtype=np.uint64)
    buffer = np.empty(8, dtype=np.uint8)
    

    # Fill hash_bytes with starting stuff 
    acc = accum_item_hash(u8(_PyHASH_XXPRIME_5), len(facts))
    for c in range(0,byte_width,8):        
        buffer0[0] = u8(accum_item_hash(acc, c))
        mx = min(byte_width, c+8)
        hash_bytes[c:mx] = buffer0.view(np.uint8)[:mx-c]

    inds = np.argsort(fact_hashes)
    c = 0
    for j, i in enumerate(inds):
        fact = facts[i]
        fact_hash = fact_hashes[i]

        nc = c+8
        mx = min(len(hash_bytes), nc)
        rest = nc-mx
        l = 8-rest

        # lower bit
        buffer[:l] = hash_bytes[c:mx]

        # upper bit that might have been cut off
        if(rest):
            buffer[l:8] = hash_bytes[:rest]

        acc = buffer.view(np.uint64)[0]
        acc = accum_item_hash(acc, j)
        acc = accum_item_hash(acc, fact_hash)

        buffer0[:] = acc
        buffer = buffer0.view(np.uint8)
        hash_bytes[c:mx] = buffer[:l]

        if(rest):
            hash_bytes[:rest] = buffer[l:8]
        
        c = (c +8) % byte_width
    return hash_bytes

@njit(cache=True)
def memset_long_hash_str(ms, n_char=30):
    byte_width = u8(np.ceil((n_char * 6) / 8.0))

    hash_bytes = memset_long_hash_bytes(ms, byte_width)
    s = byte_string_to_base64(hash_bytes, n_char)

    return s






from numba import njit
from numba.types import unicode_type, string, f8, i8
from cre.func import CREFunc
from cre.utils import _load_ptr


@CREFunc(shorthand = '{0}', no_raise=True)
def Identity(a):
    return a

@CREFunc(shorthand = '{0} == {1}', commutes=True, no_raise=True)
def Equals(a, b):
    # print("BEF", a, b)
    # print("##", a,b, a==b)
    # print("START Equals", a == b, len(a), len(b))
    return a == b

@CREFunc(shorthand = '{0} < {1}', no_raise=True)
def LessThan(a, b):
    return a < b

@CREFunc(shorthand = '{0} <= {1}', no_raise=True)
def LessThanEq(a, b):
    return a <= b


@CREFunc(shorthand = '{0} > {1}', no_raise=True)
def GreaterThan(a, b):
    return a > b

@CREFunc(shorthand = '{0} >= {1}', no_raise=True)
def GreaterThanEq(a, b):
    return a >= b

@CREFunc(shorthand = '{0} + {1}', commutes=True, no_raise=True)
def Add(a, b):
    return a + b

@CREFunc(shorthand = '{0} - {1}', no_raise=True)
def Subtract(a, b):
    return a - b

@CREFunc(shorthand = '{0} * {1}', commutes=True, no_raise=True)
def Multiply(a, b):
    return a * b

# def denom_not_zero(a,b):
#     return b != 0

@CREFunc(shorthand = '{0} / {1}')
def Divide(a, b):
    return a / b

@CREFunc(shorthand = '{0} // {1}')
def FloorDivide(a, b):
    return a // b

@CREFunc(shorthand = '{0} ** {1}')
def Power(a, b):
    return a ** b

@CREFunc(shorthand = '{0} % {1}', no_raise=True)
def Modulus(a, b):
    return a % b

@CREFunc(shorthand = '{0} < {1}', no_raise=True)
def FactIdrecsLessThan(a, b):
    return a.idrec < b.idrec

@CREFunc(shorthand = '{0} + {1}', 
    signature = unicode_type(unicode_type,unicode_type),
    commutes=False, no_raise=True)
def Concatenate(a, b):
    return a + b


@CREFunc(shorthand = '{0} == {1}', ptr_args=True)
def ObjEquals(a, b):
    '''From two head_ptrs see if the underlying pointers to objects are the same'''
    # objptr0 = _load_ptr(i8,a)
    # objptr1 = _load_ptr(i8,b)
    # return objptr0 == objptr1
    return a == b

    
@CREFunc(shorthand = '{0} == None', ptr_args=True)
def ObjIsNone(a):
    '''From a head_ptr see if the underlying object pointer is the NULL pointer'''
    # return _load_ptr(i8,a) == 0
    return a == 0


# @CREFunc(shorthand = 'float({0}', check=check_cast_float)
@CREFunc(shorthand = 'float({0})')
def CastFloat(a):
    return float(a)

# print(CastFloat(unicode_type).check('0'))
# print(CastFloat(unicode_type).check('A'))


# @CREFunc(shorthand = 'str({0}', check=check_cast_str)
@CREFunc(shorthand = 'str({0})')
def CastStr(a):
    return str(a)

# print("------------")
# print(CastStr)
# print(CastStr.__class__)
# print("------------")



from numba import types, njit, guvectorize, vectorize, prange, generated_jit
from numba.experimental import jitclass, structref
from numba import deferred_type, optional
from numba import void,b1,u1,u2,u4,u8,i1,i2,i4,i8,f4,f8,c8,c16
from numba.typed import List, Dict
from numba.core.types import DictType, ListType, unicode_type, float64, NamedTuple, NamedUniTuple, UniTuple, Array
from numba.core.extending import (
    infer_getattr,
    lower_getattr_generic,
    lower_setattr_generic,
    overload_method,
    intrinsic,
    overload,
    box,
    unbox,
    NativeValue
)
from numba.core.typing.templates import AttributeTemplate
from numba.core.datamodel import default_manager, models
from numba.core import cgutils, utils as numba_utils
from numba.experimental.structref import _Utils, imputils
from numba.typed.typedobjectutils import _nonoptional

from cre.utils import cast, _ptr_from_struct_codegen, _list_from_ptr, _cast_list
from cre.obj import member_info_type, resolve_member_id



def define_boxing(struct_type, obj_class):
    """
    
    Define the boxing & unboxing logic for `struct_type` to `obj_class`.

    Defines both boxing and unboxing.

    - boxing turns an instance of `struct_type` into a PyObject of `obj_class`
    - unboxing turns an instance of `obj_class` into an instance of
      `struct_type` in jit-code.


    Use this directly instead of `define_proxy()` when the user does not
    want any constructor to be defined.
    """
    if struct_type is types.StructRef:
        raise ValueError(f"cannot register {types.StructRef}")

    obj_ctor = obj_class._numba_box_
    struct_type._proxy_class = obj_class

    @box(struct_type)
    def box_struct_ref(typ, val, c):
        # print("BOX", typ, obj_ctor)
        """
        Convert a raw pointer to a Python int.
        """
        utils = _Utils(c.context, c.builder, typ)
        struct_ref = utils.get_struct_ref(val)
        meminfo = struct_ref.meminfo

        mip_type = types.MemInfoPointer(types.voidptr)
        boxed_meminfo = c.box(mip_type, meminfo)

        ctor_pyfunc = c.pyapi.unserialize(c.pyapi.serialize_object(obj_ctor))
        # ty_pyobj = c.pyapi.unserialize(c.pyapi.serialize_object(typ))

        res = c.pyapi.call_function_objargs(
            ctor_pyfunc, [boxed_meminfo],
        )
        c.pyapi.decref(ctor_pyfunc)
        # c.pyapi.decref(ty_pyobj)
        c.pyapi.decref(boxed_meminfo)
        return res

    @unbox(struct_type)
    def unbox_struct_ref(typ, obj, c):
        mi_obj = c.pyapi.object_getattr_string(obj, "_meminfo")

        mip_type = types.MemInfoPointer(types.voidptr)

        mi = c.unbox(mip_type, mi_obj).value

        utils = _Utils(c.context, c.builder, typ)
        struct_ref = utils.new_struct_ref(mi)
        out = struct_ref._getvalue()

        c.pyapi.decref(mi_obj)
        return NativeValue(out)


def resolve_fact_getattr_type(typ, attr):
    from cre.fact import DeferredFactRefType, Fact
    if (hasattr(typ,'spec') and attr in typ.spec):
        attrty = typ.spec[attr]['type']

        if(isinstance(attrty,DeferredFactRefType)):
            attrty = attrty.get()

        if(isinstance(attrty,Fact)):
            attrty = types.optional(attrty)
        if(isinstance(attrty, ListType)):
            if(isinstance(attrty.dtype,DeferredFactRefType)):
                attrty = ListType(attrty.dtype.get())
            attrty = types.optional(attrty)
             
        return attrty
    if attr in typ.field_dict:
        attrty = typ.field_dict[attr]
        # print("<<",attr, attrty)
        return attrty



# @intrinsic
@njit(cache=True)
def safe_get_fact_ptr(fact):
    if(fact is None):
        return 0
    else:
        return cast(_nonoptional(fact), i8)

def fact_getattr_codegen(context, builder, sig, args, attr):
    from cre.fact import Fact
    ret_type = sig.return_type
    typ, _ = sig.args
    val, _ = args
    field_type = typ.field_dict[attr]
    

    # Extract unoptional part of type
    if(isinstance(ret_type, types.Optional)):
        ret_type = ret_type.type

    utils = _Utils(context, builder, typ)
    dataval = utils.get_data_struct(val)
    ret = getattr(dataval, attr)

    option_ret_type = types.optional(ret_type)
    if(isinstance(ret_type, (Fact,))):
        # If a fact member is Null then return None
        def cast_obj(x):
            if(cast(x, i8) != 0):
                return cast(x, ret_type)
            return None
        ret = context.compile_internal(builder, cast_obj, option_ret_type(field_type,), (ret,))
        # ret_type = option_ret_type
        return ret
        
    elif(isinstance(ret_type, (ListType,))):
        # List members should always be non-null
        def cast_obj(x):
            return _cast_list(ret_type, x)
        ret = context.compile_internal(builder, cast_obj, option_ret_type(field_type,), (ret,))
        # ret_type = option_ret_type
        return ret
    else:
        
        return imputils.impl_ret_borrowed(context, builder, ret_type, ret)



@intrinsic
def fact_lower_getattr(typingctx, inst_type, attr_type):
    if (isinstance(attr_type, types.Literal) and 
        isinstance(inst_type, types.StructRef)):
        attr = attr_type.literal_value
        ret_type = resolve_fact_getattr_type(inst_type, attr)
        def codegen(context, builder, sig, args):
            return fact_getattr_codegen(context, builder, sig, args, attr)
  
        sig = ret_type(inst_type, attr_type)
        return sig, codegen

@intrinsic
def get_fact_attr_ptr(typingctx, inst_type, attr_type):
    from cre.fact import Fact
    attr = attr_type.literal_value
    def codegen(context, builder, sig, args):
        val, _ = args
        typ, _ = sig.args

        print(typ,attr)
        field_type = typ.field_dict[attr]

        if(not isinstance(field_type, (Fact))):
            raise ValueError()

        utils = _Utils(context, builder, typ)
        dataval = utils.get_data_struct(val)
        ret = getattr(dataval, attr)

        ret = _ptr_from_struct_codegen(context, builder, ret, field_type, False)

        return ret
    return u8(inst_type,attr_type), codegen

def fact_setattr_codegen(context, builder, sig, args, attr, mutability_protected=False):
    from cre.fact import Fact
    if(len(args) == 2):
        [inst_type, val_type] = sig.args
        [instance, val] = args
    else:
        [inst_type, _, val_type] = sig.args
        [instance, _, val] = args
    utils = _Utils(context, builder, inst_type)
    dataval = utils.get_data_struct(instance)

    if(mutability_protected):
        idrec = getattr(dataval, "idrec")
        # If (idec & 0xFF) != 0, throw an error 
        idrec_set = builder.icmp_unsigned('==', builder.and_(idrec, idrec.type(0xFF)), idrec.type(0))
        with builder.if_then(idrec_set):
            msg =("Facts objects are immutable once declared. Use mem.modify instead.",)
            context.call_conv.return_user_exc(builder, AttributeError, msg)

    
    field_type = inst_type.field_dict[attr]

    # print("&&", attr, field_type)
    if(isinstance(field_type, (ListType,))):
        dtype = field_type.dtype
        # if(isinstance(field_type, Fact)):
        if(isinstance(val_type, types.Optional)):
            # If list member assigned to none just instantiate an empty list
            def cast_obj(x):
                if(x is None):
                    return List.empty_list(dtype)
                return _cast_list(field_type, _nonoptional(x))
        else:
            def cast_obj(x):
                if(x is None):
                    return List.empty_list(dtype)
                return _cast_list(field_type, x)
        casted = context.compile_internal(builder, cast_obj, field_type(val_type,), (val,))

    elif(isinstance(field_type, (Fact,))):
        # If fact member assigned to none just assign to NULL pointer
        if(isinstance(val_type, types.Optional)):
            def cast_obj(x):
                if(x is None):
                    return cast(0,field_type)
                return cast(_nonoptional(x), field_type)
        else:
            def cast_obj(x):
                if(x is None):
                    return cast(0,field_type,)
                return cast(x, field_type)

        casted = context.compile_internal(builder, cast_obj, field_type(val_type,), (val,))
    else:
        casted = context.cast(builder, val, val_type, field_type)
        # incref new value
        context.nrt.incref(builder, field_type, casted)

    # pyapi = context.get_python_api(builder)
    
    # decref old value (must be last in case new value is old value)
    old_value = getattr(dataval, attr)
    context.nrt.decref(builder, field_type, old_value)
    # write new
    setattr(dataval, attr, casted)

    if(mutability_protected):
        # Make sure that hash_val is 0 to force it to be recalculated
        setattr(dataval, "hash_val", cgutils.intp_t(0))
    return dataval
        # ret = _obj_cast_codegen(context, builder, ret, field_type, ret_type, False)

    # return imputils.impl_ret_borrowed(context, builder, ret_type, ret)


@intrinsic
def fact_lower_setattr(typingctx, inst_type, attr_type, val_type):
    if (isinstance(attr_type, types.Literal) and 
        isinstance(inst_type, types.StructRef)):
        # print("BB", isinstance(inst_type, types.StructRef), inst_type, attr_type)
        
        attr = attr_type.literal_value
        def codegen(context, builder, sig, args):
            fact_setattr_codegen(context, builder, sig, args, attr)
  
        sig = types.void(inst_type, attr_type, val_type)
        # print(sig)
        return sig, codegen


@intrinsic
def fact_mutability_protected_setattr(typingctx, inst_type, attr_type, val_type):
    if (isinstance(attr_type, types.Literal) and 
        isinstance(inst_type, types.StructRef)):
        # print("BB", isinstance(inst_type, types.StructRef), inst_type, attr_type)
        
        attr = attr_type.literal_value
        def codegen(context, builder, sig, args):
            fact_setattr_codegen(context, builder, sig, args, attr, 
                mutability_protected=True)
  
        sig = types.void(inst_type, attr_type, val_type)
        # print(sig)
        return sig, codegen


def define_attributes(struct_typeclass):
    from cre.fact import Fact, base_fact_fields
    """
    Copied from numba.experimental.structref 0.51.2, but added protected mutability
    """
    # print("REGISTER FACT")
    @infer_getattr
    class StructAttribute(AttributeTemplate):
        key = struct_typeclass

        def generic_resolve(self, _typ, attr):
            typ = resolve_fact_getattr_type(_typ, attr)
            # print(attr, _typ,"!>>", typ)
            return typ

    @lower_getattr_generic(struct_typeclass)
    def struct_getattr_impl(context, builder, inst_type, val, attr):
        ret_type = resolve_fact_getattr_type(inst_type, attr)
        sig = ret_type(inst_type,types.literal(attr))
        args = (val, attr)
        return fact_getattr_codegen(context, builder, sig, args, attr)
        

    @lower_setattr_generic(struct_typeclass)
    def struct_setattr_impl(context, builder, sig, args, attr):
        fact_setattr_codegen(context, builder, sig, args, attr, mutability_protected=True)
        

        # [inst_type, val_type] = sig.args
        # [instance, val] = args
        # utils = _Utils(context, builder, inst_type)
        # dataval = utils.get_data_struct(instance)
        
        # field_type = inst_type.field_dict[attr]
        # if(isinstance(field_type, (ListType,Fact))):
        #     if(isinstance(field_type, Fact)):
        #         def cast_obj(x):
        #             if(x is None):
        #                 return _struct_from_ptr(BaseFact,0)
        #             return _cast_structref(BaseFact, _nonoptional(x))
        #     else:
        #         def cast_obj(x):
        #             return _cast_structref(field_type,x) 

        #     casted = context.compile_internal(builder, cast_obj, field_type(val_type,), (val,))
        # else:
        #     casted = context.cast(builder, val, val_type, field_type)

        # pyapi = context.get_python_api(builder)

        
            
        # # read old
        # old_value = getattr(dataval, attr)
        # # incref new value
        # context.nrt.incref(builder, field_type, casted)
        # # decref old value (must be last in case new value is old value)
        # context.nrt.decref(builder, field_type, old_value)
        # # write new
        # setattr(dataval, attr, casted)


def _register_fact_structref(fact_type):
    from cre.obj import CREObjModel
    if fact_type is types.StructRef:
        raise ValueError(f"cannot register {types.StructRef}")
    default_manager.register(fact_type, CREObjModel)
    define_attributes(fact_type)
    return fact_type


# @intrinsic
# def _fact_get_chr_mbrs_infos(typingctx, fact_type):
#     from cre.fact import base_fact_fields, base_fact_field_dict

#     t_ids = [_resolve_t_id_helper(x) for a,x in fact_type._fields if a not in base_fact_fields]
#     member_infos_out_type = types.UniTuple(member_info_type, len(t_ids))


#     def codegen(context, builder, sig, args):
#         [fact,] = args
#         utils = _Utils(context, builder, fact_type)

#         baseptr = utils.get_data_pointer(fact)
#         baseptr_val = builder.ptrtoint(baseptr, cgutils.intp_t)
#         dataval = utils.get_data_struct(fact)

#         member_infos = []
#         i = 0
#         for attr, typ in fact_type._fields:
#             # print("<<", attr, base_fact_field_dict)
#             if(attr not in base_fact_field_dict):
#                 index_of_member = datamodel.get_field_position(attr)
#                 member_ptr = builder.gep(baseptr, [cgutils.int32_t(0), cgutils.int32_t(index_of_member)], inbounds=False)
#                 member_ptr = builder.ptrtoint(member_ptr, cgutils.intp_t)
#                 offset = builder.trunc(builder.sub(member_ptr, baseptr_val), cgutils.ir.IntType(16))
#                 t_id = context.get_constant(u2, t_ids[i])
#                 member_infos.append(context.make_tuple(builder, member_info_type, (t_id, offset)))
#                 i += 1

#         ret = context.make_tuple(builder,member_infos_out_type, member_infos)
#         return ret

#     sig = member_infos_out_type(fact_type_ref)
#     return sig, codegen





@intrinsic
def _fact_get_chr_mbrs_infos(typingctx, fact_type):
    from cre.context import CREContext
    from cre.fact import base_fact_field_dict
    
    '''get the base address of the struct pointed to by structref 'inst' '''
    
    # members_type = [v for k,v in fact_type._fields if k == 'members'][0]
    context = CREContext.get_default_context()
    chr_mbr_fields = [(a,x) for a,x in fact_type._fields if a not in base_fact_field_dict and a != "chr_mbrs_infos"]
    t_ids = [context.get_t_id(_type=x)  for a,x in chr_mbr_fields]
    mbr_ids = [resolve_member_id(x) for a,x in chr_mbr_fields]

    # count = members_type.count
    member_infos_out_type = types.UniTuple(member_info_type, len(t_ids))

    
    def codegen(context, builder, sig, args):
        [fact,] = args

        utils = _Utils(context, builder, fact_type)

        baseptr = utils.get_data_pointer(fact)
        baseptr_val = builder.ptrtoint(baseptr, cgutils.intp_t)
        dataval = utils.get_data_struct(fact)
        # index_of_members = dataval._datamodel.get_field_position("members")

        member_infos = []
        i = 0
        for attr, typ in chr_mbr_fields:
            #Get the offset
            index_of_member = dataval._datamodel.get_field_position(attr)
            member_ptr = builder.gep(baseptr, [cgutils.int32_t(0), cgutils.int32_t(index_of_member)], inbounds=True)
            member_ptr = builder.ptrtoint(member_ptr, cgutils.intp_t)
            offset = builder.trunc(builder.sub(member_ptr, baseptr_val), cgutils.ir.IntType(16))

            t_id = context.get_constant(u2, t_ids[i])
            mbr_id = context.get_constant(u2, mbr_ids[i])
            member_infos.append(context.make_tuple(builder, member_info_type, (t_id, mbr_id, offset) ))
            i += 1
        ret = context.make_tuple(builder,member_infos_out_type, member_infos)
        return ret

    sig = member_infos_out_type(fact_type)
    return sig, codegen





import operator
import numpy as np
import numba
from numba import types, njit, i8, u8, i4, u1,u2,u4,  i8, literally, generated_jit, objmode
from numba.typed import List
from numba.core.types import ListType, unicode_type, void, Tuple
from numba.experimental import structref
from numba.experimental.structref import new, define_attributes, _Utils
from numba.extending import SentryLiteralArgs, lower_cast, overload_method, intrinsic, overload_attribute, intrinsic, lower_getattr_generic, overload, infer_getattr, lower_setattr_generic
from numba.core.typing.templates import AttributeTemplate
from cre.context import cre_context, get_cre_context_data
from cre.structref import define_structref, define_boxing, define_structref_template, CastFriendlyStructref
from cre.fact import define_fact, BaseFact, cast_fact, DeferredFactRefType, Fact, _standardize_type
from cre.utils import cast, PrintElapse, ptr_t, decode_idrec, lower_getattr,  lower_setattr, lower_getattr, _decref_ptr, _incref_ptr, _incref_structref, _ptr_from_struct_incref
from cre.utils import assign_to_alias_in_parent_frame, encode_idrec, _obj_cast_codegen
from cre.vector import VectorType
from cre.obj import cre_obj_field_dict,CREObjType, CREObjTypeClass, CREObjProxy, set_chr_mbrs
from cre.type_conv import ptr_to_var_name
# from cre.predicate_node import BasePredicateNode,BasePredicateNodeType, get_alpha_predicate_node_definition, \
# get_beta_predicate_node_definition, deref_attrs, define_alpha_predicate_node, define_beta_predicate_node, AlphaPredicateNode, BetaPredicateNode
from numba.core import imputils, cgutils
from numba.core.datamodel import default_manager, models
from numba.experimental.structref import StructRefProxy


from operator import itemgetter
from copy import copy
from os import getenv
from cre.utils import deref_info_type, DEREF_TYPE_ATTR, DEREF_TYPE_LIST, listtype_sizeof_item, _obj_cast_codegen
from cre.core import T_ID_VAR, register_global_default
# import inspect



var_fields_dict = {
    **cre_obj_field_dict,
    # The pointer of the Var instance before any attribute selection
    #   e.g. if '''v = Var(Type); v_b = v.B;''' then v_b.base_ptr = &v
    'base_ptr' : i8,

    # A recounted reference to the base_ptr (only nonzero on derefs i.e. v.B)
    'base_ptr_ref' : ptr_t,

    # A pointer to the Var instance which is the NOT() of this Var
    'conj_ptr' : i8,

    # The name of the Var 
    'alias_' : unicode_type,
    'deref_attrs_str' : types.optional(unicode_type),

    # The byte offsets for each attribute relative to the previous resolved
    #  fact. E.g.  if attr "B" is at offset 10 in the type assigned to "B"
    #  then v.B.B.deref_infos = [10,10]
    'deref_infos': deref_info_type[::1],

    # The t_ids of the fact that the base Var is meant to match.
    'base_t_id' : u2,

    'head_t_id' : u2,
    # If true then instead of testing for the existence of the Var
    #  we test that the Var does not exist.
    'is_not' : u1,
    # 'base_type_name': unicode_type,
    # 'head_type_name': unicode_type,
    'base_type': types.Any,
    'head_type': types.Any,

    # # Put the hashes of the head  
    # 'literal_base_hash' : types.Any,
    # 'literal_head_hash' : types.Any,
}

class VarTypeClass(CREObjTypeClass):
    t_id = T_ID_VAR
    type_cache = {}

    def __new__(cls, base_type=None, head_type=None):
        print(cls, base_type, head_type)
        if(head_type is None): head_type = base_type
        
        unq_tup = (base_type, head_type)

        self = cls.type_cache.get(unq_tup, None)
        if(self is not None):
            return self

        self = super().__new__(cls)
        self.base_type = base_type        
        self.head_type = head_type
        cls.type_cache[unq_tup] = self

        if(base_type is None or head_type is None):
            field_dict = {**var_fields_dict}
        else:
            field_dict = {**var_fields_dict,
                "base_type" : types.TypeRef(base_type),
                "head_type" : types.TypeRef(head_type)
            }
        types.StructRef.__init__(self,[(k,v) for k,v in field_dict.items()])
        
        # print(self.field_dict)
        self.name = repr(self)
        return self

    def __init__(self,*args,**kwargs):
        pass

    def __str__(self):
        if(self.base_type is None and self.head_type is None):
            return f"VarType"
        elif(self.base_type == self.head_type):
            return f"VarType[{self.base_type}]"
        else:
            return f"VarType[{self.base_type}->{self.head_type}]"
    __repr__ = __str__

# @lower_cast(VarTypeClass, CREObjType)
# def upcast(context, builder, fromty, toty, val):
#     return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)



# Manually register the type to avoid automatic getattr overloading 
default_manager.register(VarTypeClass, models.StructRefModel)

VarType = VarTypeClass()
register_global_default("Var", VarType)

# Allow typed Var instances to be upcast to VarType
@lower_cast(VarTypeClass, VarType)
def upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)

class Var(StructRefProxy):
    t_id = T_ID_VAR
    def __new__(cls, typ, alias="", skip_assign_alias=False):
        # if(not isinstance(typ, types.StructRef)): typ = typ.fact_type
        typ = _standardize_type(typ, cre_context())
        # base_type_name = str(typ)
        # print(base_type_name)
                
        base_t_id = cre_context().get_t_id(_type=typ)

        if(getenv("CRE_SPECIALIZE_VAR_TYPE",default=False)):
            raise ValueError("THIS SHOULDN'T HAPPEN")
            typ_ref = types.TypeRef(typ)
            struct_type = VarTypeClass(typ_ref,typ_ref)
            st = var_ctor(struct_type, base_t_id, alias)
        else:
            st = var_ctor_generic(base_t_id, alias)
        
        st._base_type = typ
        st._head_type = typ
        st._derefs_str = ""

        # if(alias):
            # import inspect, ctypes
            # if(alias is not None): 
            #     # Binds this instance globally in the calling python context 
            #     #  so that it is bound to a variable named whatever alias was set to
            #     # print(inspect.stack()[2][0].f_locals)
            #     # Get the calling frame
            #     frame = inspect.stack()[2][0] 
            #     # Assign the Var to it's alias
            #     frame.f_locals[alias] = st
            #     # Update locals()
            #     ctypes.pythonapi.PyFrame_LocalsToFast(ctypes.py_object(frame), ctypes.c_int(1))
            # # assign_to_alias_in_parent_frame(st, alias)

        return st
        
    def _handle_deref(self, attr_or_ind):
        '''Helper function that... '''
        
        # assert(isinstance(self.base_type,Fact))
        # with PrintElapse("Startup"):
        base_type = self.base_type
        base_type_name = str(base_type)
        # print("<<",attr_or_ind)
        _derefs_str = self.get_derefs_str()
        # print("_derefs_str", _derefs_str, type(_derefs_str))
        if(isinstance(attr_or_ind, str)):
            # ATTR case
            curr_head_type = self.head_type
            attr = attr_or_ind
            # fd = curr_head_type.field_dict
            # print("<<", attr_or_ind)
            head_type = curr_head_type.clean_spec[attr]['type']
            if(isinstance(head_type, DeferredFactRefType)):
                head_type = cre_context().name_to_type[head_type._fact_name]
            a_id = curr_head_type.get_attr_a_id(attr) #list(fd.keys()).index(attr)
            offset = curr_head_type.get_attr_offset(attr)#curr_head_type._attr_offsets[a_id]
            deref_info_type = DEREF_TYPE_ATTR
            # _derefs_str += f".{attr}"
        else:
            # LIST case
            assert isinstance(self.head_type, ListType), \
                f'__getitem__() not supported for Var with head_type {type(self.head_type)}'

            head_type = self.head_type.item_type

            attr = str(attr_or_ind)
            a_id = u4(attr_or_ind)
            print(self.head_type)
            item_size = listtype_sizeof_item(self.head_type)
            offset = int(attr_or_ind)*item_size
            # print(int(attr_or_ind), item_size)
            deref_info_type = DEREF_TYPE_LIST
            # _derefs_str += f"[{attr_or_ind}]"


        # head_t_id = getattr(head_type, "t_id", -1)
        # head_type_name = str(head_type)

        # with PrintElapse("new"):
        if(getenv("CRE_SPECIALIZE_VAR_TYPE",default=False)):

            if(deref_info_type == DEREF_TYPE_ATTR):
                struct_type = VarTypeClass(base_type, head_type)
            else:
                raise NotImplemented("Haven't implemented getitem() when CRE_SPECIALIZE_VAR_TYPE=true.")
            new_var = var_append_deref(self, attr_or_ind)#self, attr, a_id, offset, head_type_name, t_id, deref_info_type)
        else:
            head_t_id = cre_context().get_t_id(_type=head_type)
            # head_t_id = getattr(head_type, "t_id", -1)
            # print("type, a_id, offset, head_t_id", deref_info_type, a_id, offset, head_t_id)
            new_var = generic_var_append_deref(self, a_id, offset, head_t_id, typ=deref_info_type)
            # new = generic_var_append_deref(self, attr, a_id, offset, head_type_name, t_id, deref_info_type)
            # struct_type = VarType
        #CHECK THAT PTRS ARE SAME HERE
            

            
        new_var._base_type = base_type
        new_var._head_type = head_type
        # new_var._derefs_str = _derefs_str

        # new = var_ctor(struct_type, str(type_name), var_get_alias(self))
        # var_memcopy(self, new)
        # var_append_deref(new, attr, offset)
        return new_var





    def __getattr__(self, attr):
        if(attr in ['_head_type', '_base_type','_derefs_str', '_deref_attrs']): return None
        if(attr == 'deref_attrs_str'):
            if('_deref_attrs_str' not in self.__dict__):
                self._deref_attrs_str = "".join(
                    [f"[{a}]" if a.isdigit() else "." + a for a in self.deref_attrs]
                )
            return self._deref_attrs_str
        elif(attr == 'deref_attrs'):
            if('_deref_attrs' not in self.__dict__):
                self._deref_attrs = resolve_deref_attrs(self)
            return self._deref_attrs
        elif(attr == 'base_type'):
            # Otherwise we need to resolve the type from the current context
            if('_base_type' not in self.__dict__):
                base_type_ref = self._numba_type_.field_dict['base_type']
                if(base_type_ref != types.Any):
                    # If the Var is type specialize then grab its base_type
                    self._base_type = base_type_ref.instance_type 
                else:
                    context = cre_context()
                    self._base_type = context.get_type(t_id=self.base_t_id)
            return self._base_type
        elif(attr == 'head_type' or attr == 'return_type'):
            if('_head_type' not in self.__dict__):
                head_type_ref = self._numba_type_.field_dict['head_type']
                if(head_type_ref != types.Any):
                    self._head_type = head_type_ref.instance_type 
                else:
                    context = cre_context()
                    self._head_type = context.get_type(t_id=self.head_t_id)
            return self._head_type
        elif(attr == 'is_not'):
            return var_get_is_not(self)
        
        #     return var_get_deref_attrs(self)
        elif(attr == 'deref_infos'):
            return var_get_deref_infos(self)
        elif(attr == 'alias'):
            return var_get_alias(self)
        elif(attr == 'base_t_id'):
            return var_get_base_t_id(self)
        elif(attr == 'head_t_id'):
            return var_get_head_t_id(self)
        elif(attr == 'base_ptr'):
            return var_get_base_ptr(self)
        elif(True): 
            return self._handle_deref(attr)

    def __getitem__(self,ind):
        return self._handle_deref(ind)

    def __str__(self):
        if(self.alias == ""):
            return self.__repr__()
        if(self.is_not):
            return f'NOT({self.alias}){self.deref_attrs_str}'
        else:
            return f'{self.alias}{self.deref_attrs_str}'

    def __repr__(self):
        prefix = "NOT" if(self.is_not) else "Var"
        base_name = self.base_type._fact_name if hasattr(self.base_type,'_fact_name') else str(self.base_type)
        if(self.alias != ""):
            base = f'{prefix}({base_name},{self.alias!r})'
        else: 
            base = f'{prefix}({base_name})'
        # print(self.deref_attrs)
        return f'{base}{self.deref_attrs_str}'

    def _cmp_helper(self,op_str,other,negate):
        check_legal_cmp(self, op_str, other)
        opt_str = types.literal(types.unliteral(op_str))
        # print(other)
        
        if(not isinstance(other,(VarTypeClass,Var))):
            # print("other",other)
            if(isinstance(other,(bool,))): other = int(other)
            return var_cmp_alpha(self,op_str,other, negate)
        else:
            return var_cmp_beta(self,op_str,other, negate)
    

    def __lt__(self, other): 
        from cre.default_funcs import LessThan, FactIdrecsLessThan
        if(isinstance(other,Var) and isinstance(other.head_type,Fact)):
            return FactIdrecsLessThan(self,other)
        else:
            return LessThan(self, other)
    def __le__(self, other): 
        from cre.default_funcs import LessThanEq
        return LessThanEq(self, other)
            
    def __gt__(self, other): 
        from cre.default_funcs import GreaterThan, FactIdrecsLessThan
        if(isinstance(other,Var) and isinstance(other.head_type,Fact)):
            return FactIdrecsLessThan(other,self)
        else:
            return GreaterThan(self, other)

    def __ge__(self, other):
        from cre.default_funcs import GreaterThanEq
        return GreaterThanEq(self, other)
    def __eq__(self, other): 
        from cre.default_funcs import Equals, ObjEquals, ObjIsNone
        from cre.conditions import cre_func_to_cond

        # with PrintElapse("new_ptr_op"):
        #     npo = ObjIsNone(self)

        # with PrintElapse("op_to_cond"):
        #     op_to_cond(npo)
        
        if(other is None):
            # with PrintElapse("new_ObjIsNone"):
            return cre_func_to_cond(ObjIsNone(self))
        if(isinstance(other,Var) and isinstance(other.head_type,Fact)):
            # print("ObjEquals")
            # with PrintElapse("new_ObjEquals"):
            return cre_func_to_cond(ObjEquals(self,other))
        # print("Equals")
        return Equals(self, other)
    def __ne__(self, other): 
        return ~(self == other)
        # from cre.default_funcs import Equals, ObjEquals, ObjIsNone
        # if(other is None):
        #     return ~ObjIsNone(other)
        # if(isinstance(other,Var) and isinstance(other.head_type,Fact)):
        #     return ~ObjEquals(self,other)
        # return ~Equals(self, other)

    def __add__(self, other):
        from cre.default_funcs import Add
        return Add(self, other)

    def __radd__(self, other):
        from cre.default_funcs import Add
        return Add(other, self)

    def __sub__(self, other):
        from cre.default_funcs import Subtract
        return Subtract(self, other)

    def __rsub__(self, other):
        from cre.default_funcs import Subtract
        return Subtract(other, self)

    def __mul__(self, other):
        from cre.default_funcs import Multiply
        return Multiply(self, other)

    def __rmul__(self, other):
        from cre.default_funcs import Multiply
        return Multiply(other, self)

    def __truediv__(self, other):
        from cre.default_funcs import Divide
        return Divide(self, other)

    def __rtruediv__(self, other):
        from cre.default_funcs import Divide
        return Divide(other, self)

    def __floordiv__(self, other):
        from cre.default_funcs import FloorDivide
        return FloorDivide(self, other)

    def __rfloordiv__(self, other):
        from cre.default_funcs import FloorDivide
        return FloorDivide(other, self)

    def __pow__(self, other):
        from cre.default_funcs import Power
        return Power(self, other)

    def __rpow__(self, other):
        from cre.default_funcs import Power
        return Power(other, self)

    def __mod__(self, other):
        from cre.default_funcs import Modulus
        return Modulus(other, self)

    def __and__(self, other):
        from cre.conditions import conditions_and, cre_func_to_cond
        from cre.func import CREFunc
        if(isinstance(other,CREFunc)):
            other = cre_func_to_cond(other)
        out = conditions_and(self, other)
        return out

    def __or__(self, other):
        from cre.conditions import conditions_or, cre_func_to_cond
        from cre.func import CREFunc
        if(isinstance(other,CREFunc)): other = cre_func_to_cond(other)
        return conditions_or(self, other)

    def __invert__(self):
        from cre.conditions import _var_NOT
        return _var_NOT(self)
    
    # Explicitly defining these allows for pickling w/o invoking __getattr__()
    def __getstate__(self):
        return self.__dict__
    def __setstate__(self, d):
        self.__dict__ = d

    def __hash__(self):
        from cre.dynamic_exec import _cre_obj_hash
        # ptr = 0#get_var_ptr(self)
        return _cre_obj_hash(self)

    def get_base_ptr(self):
        return var_get_base_ptr(self)

    def get_ptr(self):
        return get_var_ptr(self)

    def get_ptr_incref(self):
        return get_var_ptr_incref(self)

    def get_derefs_str(self):
        if(self._derefs_str is None):
            self._derefs_str = "".join([f"[{x}]" if x.isdigit() else f".{x}" for x in self.deref_attrs])#str_var_derefs(self)
        return self._derefs_str

    @classmethod
    def from_ptr(cls, ptr):
        return var_from_ptr(ptr)

    def copy(self):
        return var_copy(self)

    def __copy__(self):
        return var_copy(self)

    def with_alias(self, alias):
        return var_with_alias(self, alias)


# Manually define the boxing to avoid constructor overloading
define_boxing(VarTypeClass, Var)

#### Get Attribute Overloading ####

@infer_getattr
class StructAttribute(AttributeTemplate):
    key = VarTypeClass
    def generic_resolve(self, typ, attr):
        if(attr == "alias"): return unicode_type
        from numba.cpython.hashing import _Py_hash_t
        if(attr == "__hash__"): return types.FunctionType(_Py_hash_t(CREObjType,))
        
        
        if attr in typ.field_dict:

            attrty = typ.field_dict[attr]
            # print("is attr", attrty)
            return attrty
        # print("AAAAA", typ.field_dict)

        # print("<<", attr, typ.field_dict)
        head_type = typ.head_type
        #TODO Should check that alld subtype references are valid
        if(not hasattr(head_type,'field_dict')):
            raise AttributeError(f"Cannot dereference attribute '{attr}' of {typ}.")

        # base_type = typ.field_dict['base_type']
        if(attr in head_type.field_dict):
            # head_type = types.TypeRef(head_type.field_dict[attr])
            # field_dict = {
            #     **var_fields_dict,
            #     **{"base_type" : base_type,
            #      "head_type" : head_type}
            # }
            attrty = VarTypeClass(typ.base_type, head_type.field_dict[attr])
            print("NEW", attrty)
            # attrty = VarTypeClass([(k,v) for k,v, in field_dict.items()])
            return attrty
        else:
            raise AttributeError(f"Var[{base_type}] has no attribute '{attr}'")

@lower_getattr_generic(VarTypeClass)
def var_getattr_impl(context, builder, typ, val, attr):
    
    #If the attr is 'alias' then retrieve the base var's alias
    if(attr == "alias"):
        st = cgutils.create_struct_proxy(typ)(context, builder, value=val)._getvalue()

        def get_alias(self):
            if(self.base_ptr != cast(self, i8)):
                base = cast(self.base_ptr, VarType)
            else:
                base = cast(self, VarType)
            return base.alias_

        ret = context.compile_internal(builder, get_alias, unicode_type(typ,), (st,))
        # context.nrt.incref(builder, unicode_type, ret)
        return ret

    #If the attribute is one of the var struct fields then retrieve it.
    elif(attr in var_fields_dict):
        # p
        # print("GETATTR", attr)
        utils = _Utils(context, builder, typ)
        dataval = utils.get_data_struct(val)
        ret = getattr(dataval, attr)
        fieldtype = typ.field_dict[attr]
        return imputils.impl_ret_borrowed(context, builder, fieldtype, ret)

    #Otherwise return a new instance with a new 'attr' and 'offset' append 
    else:
        # print("APPEND")
        base_type = typ.field_dict['base_type'].instance_type 
        head_type = typ.field_dict['head_type'].instance_type.clean_spec[attr]['type']
        # head_type = typ.field_dict['head_type'].instance_type 
        head_type_name = str(head_type)
        # print(">>", head_type_name)
        fd = base_type.field_dict
        a_id = base_type.get_attr_a_id(attr)#list(fd.keys()).index(attr)
        offset = base_type.get_attr_offset(attr)#base_type._attr_offsets[list(fd.keys()).index(attr)]
        ctor = cgutils.create_struct_proxy(typ)
        st = ctor(context, builder, value=val)._getvalue()
        t_id = getattr(head_type, "t_id", -1)

        new_var_type = VarTypeClass(base_type,head_type)

        def new_var_and_append(self):
            return var_append_deref(self, attr)
            # st = new(typ)
            # var_memcopy(self,st)
            # var_append_deref(st,attr,offset)
        # return st

        ret = context.compile_internal(builder, new_var_and_append, new_var_type(typ,), (st,))
        context.nrt.incref(builder, typ, ret)
        return ret



def resolve_deref_attrs(self):
    '''Gets the chain of attribute strings for a Var e.g. x.A.B.C -> ['A','B','C']
    '''
    context = cre_context()
    deref_infos = var_get_deref_infos(self)
    deref_attrs = []
    typ = self.base_type
    for i, x in enumerate(deref_infos):
        if(isinstance(typ, ListType)):
            deref_attrs.append(str(x['a_id']))
        else:
            try:
                deref_attrs.append(typ.get_attr_from_a_id(x['a_id']))
            except IndexError as e:
                # print(self.base_t_id)
                chain_so_far = ''.join([f'[{a}]' if a.isdigit() else f'.{a}' for a in deref_attrs])
                raise ValueError(
                "Could resolve next attribute in chain after Var(" +\
                f"{typ},{self.alias!r}){chain_so_far}. No a_id {x['a_id']}." +\
                f"Chain: {deref_infos}"
                )

        typ = context.get_type(t_id=x['t_id'])
    return deref_attrs

# ### Methods that require python interpreter ### 
# # Note: could get around this is there was some way to load cre_context.context_data
# #  as some kind of global variable within the numba runtime

# @generated_jit
# @overload_method(VarTypeClass, "get_head_type_name")
# def get_head_type_name(self):
#     def impl(self):
#         # print("GET HEAD")
#         with objmode(head_type_name=unicode_type):
#             context = cre_context()
#             head_type_name = str(context.get_type(t_id=self.head_t_id))
#         return head_type_name
#     return impl

# @generated_jit
# @overload_method(VarTypeClass, "get_base_type_name")
# def get_base_type_name(self):
#     def impl(self):
#         with objmode(base_type_name=unicode_type):
#             context = cre_context()
#             base_type = context.get_type(t_id=self.base_t_id)
#             base_type_name = getattr(base_type,'_fact_name',str(base_type))
#         return base_type_name
#     return impl


@njit(ListType(unicode_type)(VarType), cache=True)
def get_deref_attrs(self):
    deref_attrs = List.empty_list(unicode_type)
    context_data = get_cre_context_data()
    t_id = self.base_t_id
    for di in self.deref_infos:
        if(di.type == DEREF_TYPE_LIST):
            deref_attrs.append(f"[{di.a_id}]")
        else:
            deref_attrs.append(context_data.attr_names[(u2(t_id), u1(di.a_id))])
        t_id = di.t_id
    for di in self.deref_infos:
        deref_attrs.append(context_data.attr_names[(u2(t_id), u1(di.a_id))])
        t_id = di.t_id
    return deref_attrs

@njit(unicode_type(VarType), cache=True)
def get_deref_attrs_str(self):
    context_data = get_cre_context_data()
    t_id = self.base_t_id
    s = ""
    for di in self.deref_infos:
        if(di.type == DEREF_TYPE_LIST):
            s += f"[{di.a_id}]"
        else:
            s += f".{context_data.attr_names[(u2(t_id), u1(di.a_id))]}"
        t_id = di.t_id
    return s


@njit(unicode_type(VarType), cache=True)
def var_str(self):
    alias = self.alias
    deref_str = get_deref_attrs_str(self)
    if(len(alias)==0):
        return f'{ptr_to_var_name(self.base_ptr)}{deref_str}'
    else:
        return f'{self.alias}{deref_str}' 

# @generated_jit(cache=True)
@overload_method(VarTypeClass, '__str__')
@overload(str)
def var_str_overload(self):
    if(not isinstance(self,VarTypeClass)): return
    def impl(self):
        return var_str(self)
    return impl






# def var_cmp_alpha(left_var, op_str, right_var,negated):
#     from cre.conditions import pt_to_cond, gen_pterm_ctor_alpha, gen_pterm_ctor_beta
#     # Treat None as 0 for comparing against a fact ref
#     print("***", isinstance(left_var.head_type, Fact),isinstance(left_var.head_type, types.StructRef), left_var.head_type)
#     if(right_var is None and isinstance(left_var.head_type, types.StructRef)): right_var = 0
#     right_var_type = types.unliteral(types.literal(right_var)) #if (isinstance(right_var, types.NoneType)) else types.int64
#     ctor = gen_pterm_ctor_alpha(left_var, op_str, right_var_type)
#     pt = ctor(left_var, op_str, right_var)
#     lbv = cast_structref(VarType,left_var)
#     return pt_to_cond(pt, lbv, None, negated)
    

# def var_cmp_beta(left_var, op_str, right_var, negated):
#     from cre.conditions import pt_to_cond, gen_pterm_ctor_alpha, gen_pterm_ctor_beta
#     ctor = gen_pterm_ctor_beta(left_var, op_str, right_var)
#     pt = ctor(left_var, op_str, right_var)
#     lbv = cast_structref(VarType,left_var)
#     rbv = cast_structref(VarType,right_var)
#     return pt_to_cond(pt, lbv, rbv, negated)


def check_legal_cmp(var, op_str, other_var):
    # if(isinstance(var,Var)): var = var
    # if(isinstance(other_var,Var)): other_var = other_var
    if(op_str != "=="):
        head_type = var.head_type
        other_head_type = None
        if(isinstance(other_var, (VarTypeClass,Var))):
            other_head_type = other_var.head_type

        # if(hasattr(head_type, '_fact_name') or hasattr(other_head_type, '_fact_name')):
        #     raise AttributeError("Inequality not valid comparitor for Fact types.")


@njit(cache=True)
def var_get_is_not(self):
    return self.is_not

# @njit(cache=True)
# def var_get_deref_attrs(self):
    # return self.deref_attrs

@njit(cache=True)
def var_get_deref_infos(self):
    return self.deref_infos

@njit(cache=True)
def var_get_alias(self):
    return self.alias

@njit(cache=True)
def var_get_base_ptr(self):
    return self.base_ptr

@njit(cache=True)
def var_get_head_t_id(self):
    return self.head_t_id

@njit(cache=True)
def var_get_base_t_id(self):
    return self.base_t_id



@njit(VarType(i8), cache=True)
def var_from_ptr(ptr):
    return cast(ptr, VarType)

@njit(cache=True)    
def get_var_ptr(self):
    return cast(self, i8)

@njit(cache=True)    
def get_var_ptr_incref(self):
    return _ptr_from_struct_incref(self)


# var_type_cache = {}
# def VarTypeClass(base_type, head_type=None):
#     if(head_type is None): head_type = base_type
#     t = (base_type, head_type)
#     if(t not in var_type_cache):
#         # print((str(t[0]),str(t[1])), t[0].t_id)
#         # d = {**var_fields_dict,**{
#         #     'base_type': types.TypeRef(base_type),
#         #     'head_type': types.TypeRef(head_type),
#         #     }}

#         struct_type = VarTypeClass(base_type, head_type)
#         var_type_cache[t] = struct_type
#         return struct_type
#     else:
#         # print("RETREIVED", (str(t[0]),str(t[1])),t[0].t_id)
#         return var_type_cache[t]

@njit(cache=True)
def var_ctor(var_struct_type, base_t_id, alias=""):
    st = new(var_struct_type)
    lower_setattr(st,'idrec', encode_idrec(T_ID_VAR,0,0xFF))
    lower_setattr(st,'is_not', u1(0))
    lower_setattr(st,'conj_ptr', i8(0))
    lower_setattr(st,'base_t_id', base_t_id)
    lower_setattr(st,'head_t_id', base_t_id)
    lower_setattr(st,'base_ptr', cast(st, i8))
    lower_setattr(st,'base_ptr_ref', ptr_t(0))
    lower_setattr(st,'alias_', alias)
    lower_setattr(st,'deref_attrs_str', None)
    lower_setattr(st,'deref_infos', np.empty(0,dtype=deref_info_type))
    # base_type_name = lower_getattr(self,"base_type_name")
    

    # st.idrec = encode_idrec(T_ID_VAR,0,0xFF)
    # st.is_not = u1(0)
    # st.conj_ptr = i8(0)
    # st.base_t_id = base_t_id
    # st.head_t_id = base_t_id
    # st.base_ptr = i8(cast(st,i8))
    # st.base_ptr_ref = ptr_t(0)
    # st.alias =  "" if(alias is  None) else alias
    # st.deref_attrs_str = None
    # # st.deref_attrs = List.empty_list(unicode_type)
    # st.deref_infos = np.empty(0,dtype=deref_info_type)
    return st


@njit(VarType(u2, unicode_type), cache=True)
def var_ctor_generic(base_t_id, alias):
    return var_ctor(VarType, base_t_id, alias)


@overload(Var)
def overload_Var(typ,alias=""):
    _typ = typ.instance_type
    struct_type = VarTypeClass(_typ,_typ)
    base_t_id = cre_context().get_t_id(_type=_typ)
    # print("@@ IMPL VAR :: ", _typ, base_t_id)
    def impl(typ, alias=""):
        return var_ctor(struct_type, base_t_id, alias)

    return impl


#### getattr and dereferencing ####

# @njit(types.void(VarType, VarType),cache=True)
# def var_memcopy(self,st):
#     # new_deref_attrs = List.empty_list(unicode_type)
#     # new_deref_infos = np.empty(len(),dtype=deref_info_type)
#     # for x in lower_getattr(self,"deref_attrs"):
#     #     new_deref_attrs.append(x)
#     # old_deref_infos
#     # for i,y in enumerate(lower_getattr(self,"deref_infos")):
#     #     new_deref_infos[i] = y
    

# @generated_jit(cache=True)
# def var_append_deref(self,deref):

# ------------------------
# : Copying, extending, and renaming Vars
#   Note: We use lower_get/setattr to get around the fact that Var().whatever 
#     creates a new var extended with a dereference.        

@njit(VarType(VarType), cache=True)
def var_copy(self):
    st = new(VarType)

    deref_infos = lower_getattr(self,"deref_infos").copy()
    lower_setattr(st,'deref_infos', deref_infos)

    if(len(deref_infos) == 0):
        # If we are copying a base Var then it is its own base.
        lower_setattr(st,'base_ptr', cast(st, i8))
        lower_setattr(st,'base_ptr_ref', ptr_t(0))
        lower_setattr(st,'alias_', lower_getattr(self,"alias_"))
    else:
        # If we are copying a Var with derefs then we need to borrow a
        #  reference to the original base.
        base_ptr_ref = lower_getattr(self, "base_ptr_ref")
        if(not base_ptr_ref == 0):
            _incref_ptr(base_ptr_ref)
        lower_setattr(st,'base_ptr_ref', base_ptr_ref)
        lower_setattr(st,'base_ptr', lower_getattr(self,"base_ptr"))
    
    lower_setattr(st,'idrec', lower_getattr(self,"idrec"))
    lower_setattr(st,'is_not', lower_getattr(self,"is_not"))
    lower_setattr(st,'base_t_id', lower_getattr(self,"base_t_id"))
    lower_setattr(st,'head_t_id', lower_getattr(self,"head_t_id"))
    lower_setattr(st,'deref_attrs_str', lower_getattr(self,"deref_attrs_str"))
    return st    

@njit(VarType(VarType, unicode_type), cache=True)
def var_with_alias(self, alias):
    st = var_copy(self)
    if(len(self.deref_infos) == 0):
        lower_setattr(st, "alias_", alias)
    else:
        # If realiasing a Var with derefs then we need to make a new base
        old_base = cast(lower_getattr(self,"base_ptr"), VarType)
        new_base = var_copy(old_base)
        lower_setattr(new_base, "alias_", alias)
        # Borrow a reference to the new base
        lower_setattr(st, "base_ptr_ref", _ptr_from_struct_incref(new_base))
        lower_setattr(st, "base_ptr", cast(new_base,i8))
    return st


@njit(types.void(VarType, u4, i4, u2, u1), cache=True)
def _var_append_deref(self, a_id, offset, head_t_id, typ):
    # lower_getattr(self,"deref_attrs").append(attr)
    old_deref_infos = lower_getattr(self,"deref_infos")
    L = len(old_deref_infos)
    new_deref_infos = np.empty(L+1,dtype=deref_info_type)
    new_deref_infos[:L] = old_deref_infos
    if(typ == DEREF_TYPE_ATTR):
        new_deref_infos[L].type = u1(DEREF_TYPE_ATTR)
    elif(typ == DEREF_TYPE_LIST):
        new_deref_infos[L].type = u1(DEREF_TYPE_LIST)

    new_deref_infos[L].a_id = u4(a_id)
    new_deref_infos[L].offset = i4(offset)
    new_deref_infos[L].t_id = u2(head_t_id)

    lower_setattr(self,'deref_infos', new_deref_infos)
    lower_setattr(self,'head_t_id', head_t_id)
    # lower_setattr(self,'base_ptr_ref', ptr_t(lower_getattr(self,"base_ptr")))


@generated_jit(cache=True)
def var_append_deref(self, attr):
    SentryLiteralArgs(['attr']).for_function(var_append_deref).bind(self,attr)
    if(self is VarType):
        raise ValueError("var_append_deref() doesn't work on VarType. Use generic_var_append_deref()")

    old_var_type = self
    old_head_type = self.field_dict['head_type'].instance_type
    attr = attr.literal_value
    a_id = old_head_type.get_attr_a_id(attr)
    offset = old_head_type.get_attr_offset(attr)
    # print(old_head_type.instance_type.field_dict)
    head_type = old_head_type.field_dict[attr]
    base_type = old_var_type.field_dict['base_type'].instance_type
    head_t_id = cre_context().get_t_id(_type=head_type)
    
    var_struct_type = VarTypeClass(base_type, head_type)

    # print("CONSTR VAR", base_type, head_type, self.name)
    # print("<<", var_struct_type)
    if(isinstance(attr,int)):
        typ = DEREF_TYPE_ATTR
    else:
        typ = DEREF_TYPE_LIST

    # print("AFT")
    def impl(self, attr):
        # st = new(var_struct_type)
        # var_memcopy(cast(self, VarType),st)
        # lower_setattr(st, 'base_ptr_ref', _ptr_from_struct_incref(self))
        st = var_copy(self)
        _var_append_deref(st, a_id, offset, head_t_id, typ=DEREF_TYPE_ATTR)
        was_base = len(self.deref_infos) == 0
        if(was_base):
            # If was a base var then borrow its pointer
            lower_setattr(st, "base_ptr", cast(self, i8))
            lower_setattr(st, 'base_ptr_ref', _ptr_from_struct_incref(self))
        lower_setattr(st,'deref_attrs_str', None)
        return st
    return impl


# def deref_var(struct_type, base_var, attr, a_id, offset, head_type_name, t_id=-1, typ=DEREF_TYPE_ATTR):
#     # _incref_structref(base_var)
#     st = new(struct_type)
#     var_memcopy(base_var,st)
#     _var_append_deref(st,attr, a_id, offset, t_id, head_type_name, typ)
#     lower_setattr(st, 'base_ptr_ref', _ptr_from_struct_incref(base_var))
#     return st

@njit(VarType(VarType, u4, i8, i8, i8), cache=True)
def generic_var_append_deref(self, a_id, offset, head_t_id, typ=DEREF_TYPE_ATTR):
    # _incref_structref(base_var)
    st = var_copy(self)
    # st = new(VarType)
    was_base = len(self.deref_infos) == 0
    # var_memcopy(self, st)
    if(was_base):
        # If was a base var then borrow its pointer
        lower_setattr(st, "base_ptr", cast(self, i8))
        lower_setattr(st, 'base_ptr_ref', _ptr_from_struct_incref(self))
    _var_append_deref(st, a_id, offset, head_t_id, typ=typ)
    lower_setattr(st,'deref_attrs_str', None)
    

    return st

@njit(VarType(VarType, deref_info_type[::1]), cache=True)
def var_extend(self, deref_infos):
    st = var_copy(self)
    was_base = len(self.deref_infos) == 0
    if(was_base):
        # If was a base var then borrow its pointer
        lower_setattr(st, "base_ptr", cast(self, i8))
        lower_setattr(st, 'base_ptr_ref', _ptr_from_struct_incref(self))
    # st = new(VarType)
    # was_base = len(self.deref_infos) == 0
    # var_memcopy(self, st)
    # if(was_base):
    #     # If was a base var then borrow its pointer
    #     lower_setattr(st, 'base_ptr_ref', _ptr_from_struct_incref(self))

    new_deref_infos = np.empty(len(self.deref_infos) + len(deref_infos), dtype=deref_info_type)
    new_deref_infos[:len(self.deref_infos)] = self.deref_infos
    new_deref_infos[len(self.deref_infos):] = deref_infos
    lower_setattr(st, 'deref_infos', new_deref_infos)
    lower_setattr(st, 'head_t_id', new_deref_infos[-1].t_id)
    lower_setattr(st,'deref_attrs_str', None)
    
    return st



@lower_setattr_generic(VarTypeClass)
def struct_setattr_impl(context, builder, sig, args, attr):



    [inst_type, val_type] = sig.args
    [instance, val] = args
    utils = _Utils(context, builder, inst_type)
    dataval = utils.get_data_struct(instance)
    # cast val to the correct type
    field_type = inst_type.field_dict[attr]
    casted = context.cast(builder, val, val_type, field_type)
    # read old
    old_value = getattr(dataval, attr)
    # incref new value
    context.nrt.incref(builder, val_type, casted)
    # decref old value (must be last in case new value is old value)
    context.nrt.decref(builder, val_type, old_value)
    # write new
    setattr(dataval, attr, casted)


@njit(types.void(VarType, unicode_type), cache=True)
def var_assign_alias(var, alias):
    base = cast(var.base_ptr, VarType)
    base.alias_ = alias

#### dereferencing for py_funcs ####

@intrinsic
def _var_deref(typingctx, typ, attr_type):
    attr = attr_type.literal_value
    def codegen(context, builder, sig, args):
        impl = context.get_getattr(sig.args[0], attr)
        return impl(context, builder, sig.args[0], args[0], attr)

    sig = typ(typ,attr_type)
    return sig, codegen

@njit(cache=True)
def var_deref(self,attr):
    return _var_deref(self,literally(attr))







# from cre.fact import _fact_from_spec, _standardize_spec, _merge_spec_inheritance, \
#      define_fact, cast_fact, _cast_structref, BaseFact, BaseFactType, DeferredFactRefType
from cre.utils import _cast_structref
from cre.context import cre_context
from cre.tuple_fact import TupleFact, TF#TupleFact, assert_cre_obj, TupleFact
from cre.fact import define_fact#TupleFact, assert_cre_obj, TupleFact
# from cre.primitive import Primitive, StringPrimitiveType
from cre.obj import CREObjType
from cre.utils import PrintElapse, decode_idrec

from numba import njit, f8, u8, i8, bool_
from numba.types import unicode_type
from numba.typed import List, Dict
import pytest
import numpy as np

import cre.dynamic_exec
import pytest



@njit(cache=True)
def as_cre_obj(x):
    return _cast_structref(CREObjType,x)

# @njit(cache=True)
# def init_prim():
#     return Primitive(2)

@njit(bool_(CREObjType, CREObjType),cache=True)
def eq(a,b):
    return a == b


def test_init():
    @njit(cache=True)
    def make_it():
        return TF("HI",TF("HI",2))

    tf = make_it.py_func()
    tf = make_it()








def test_eq():
    # hi = Primitive("HI")
    # a1 = as_cre_obj(Primitive(True))
    # a2 = as_cre_obj(Primitive(True))
    # b = as_cre_obj(Primitive(False))
    
    # assert eq(a1, a2) == True
    # assert eq(a1, b) == False

    # a1 = as_cre_obj(Primitive(2))
    # a2 = as_cre_obj(Primitive(2))
    # b = as_cre_obj(Primitive(3))

    # assert eq(a1, a2) == True
    # assert eq(a1, b) == False

    # a1 = as_cre_obj(Primitive(1.0))
    # a2 = as_cre_obj(Primitive(1.0))
    # b = as_cre_obj(Primitive(2.0))
    
    # assert eq(a1, a2) == True
    # assert eq(a1, b) == False

    # a1 = as_cre_obj(Primitive("A"))
    # a2 = as_cre_obj(Primitive("A"))
    # b = as_cre_obj(Primitive("B"))
    
    # assert eq(a1, a2) == True
    # assert eq(a1, b) == False

    a1 = TF("HI",TF("HI",2))
    a2 = TF("HI",TF("HI",2))
    b1 = TF("HI",TF("HI",3))
    b2 = TF("HI",TF("HO",2))
    b3 = TF("HO",TF("HI",2))
    b4 = TF("HO",TF("HI",2, 0))

    print(decode_idrec(a1.idrec))
    print(decode_idrec(a2.idrec))
    print(decode_idrec(b1.idrec))
    print(decode_idrec(b2.idrec))
    print(decode_idrec(b3.idrec))
    print(decode_idrec(b4.idrec))

    # print(eq(a1, a2))
    assert eq(a1, a2)
    assert not eq(a1, b1)
    assert not eq(a1, b2)
    assert not eq(a1, b3)
    assert not eq(a1, b4)

@njit(cache=True)
def hsh(x):
    return hash(x)


def test_hash():
    # a1 = as_cre_obj(Primitive(True))
    # print(hsh(a1))    

    # a1 = as_cre_obj(Primitive(2))
    # print(hsh(a1))

    # a1 = as_cre_obj(Primitive(1.0))
    # print(hsh(a1))
    
    # a1 = as_cre_obj(Primitive("A"))
    # print(hsh(a1))

    # a1 = as_cre_obj(TupleFact("HI",2))
    # print(hsh(a1))

    a1 = TF("HI",TF("HI",2))
    a2 = TF("HI",TF("HI",2))
    b1 = TF("HI",TF("HI",3))
    b2 = TF("HI",TF("HO",2))
    b3 = TF("HO",TF("HI",2))
    b4 = TF("HO",TF("HI",2, 0))

    print(hsh(a1),hsh(a2))
    print(hsh(a1),hsh(b1))
    print(hsh(a1),hsh(b2))
    print(hsh(a1),hsh(b3))
    print(hsh(a1),hsh(b4))


    assert hsh(a1) == hsh(a2)
    assert hsh(a1) != hsh(b1)
    assert hsh(a1) != hsh(b2)
    assert hsh(a1) != hsh(b3)
    assert hsh(a1) != hsh(b4)

def test_hash_obj_builtin_members():
    from cre.default_funcs import Equals
    from cre.var import Var
    eq_f8 = Equals(f8,f8)
    eq_str = Equals(unicode_type,unicode_type)
    a_f8 = Var(f8,'a')
    b_f8 = Var(f8,'b')
    a_str = Var(unicode_type,'a')
    b_str = Var(unicode_type,'b')
    
    a1 = TF(eq_f8, a_f8, b_f8)
    a2 = TF(eq_f8, a_f8, b_f8)
    b1 = TF(eq_str, a_str, b_str)

    a1_hash = hsh(a1)
    a2_hash = hsh(a1)

    print(hsh(a1), hsh(a2))
    print(hsh(a1), hsh(b1))

    assert hsh(a1) == hsh(a2)
    assert hsh(a1) != hsh(b1)

def test_eq_obj_builtin_members():
    from cre.default_funcs import Equals
    from cre.var import Var
    eq_f8 = Equals(f8,f8)
    eq_str = Equals(unicode_type,unicode_type)
    a_f8 = Var(f8,'a')
    b_f8 = Var(f8,'b')
    a_str = Var(unicode_type,'a')
    b_str = Var(unicode_type,'b')
    
    a1 = TF(eq_f8, a_f8, b_f8)
    a2 = TF(eq_f8, a_f8, b_f8)
    b1 = TF(eq_str, a_str, b_str)

    a1_hash = hsh(a1)
    a2_hash = hsh(a1)

    assert eq(a1, a2)
    assert not eq(a1, b1)
    

    # print(a1, a2)


N = 10000


def test_str():
    with cre_context("test_str"):
        a1 = TF("HI", TF("HI", 2))
        # assert str(at) == 'TF("HI", TF("HI", 2))'
        print(a1)

        a1_typ = TF(unicode_type, TF(unicode_type, i8))
        print(a1_typ)

        # assert str(a1_typ) == 'TF(unicode_type, TF(unicode_type, i8))'

        BOOP = define_fact("BOOP", {"A" : a1_typ, "B" : "number"})

        b = BOOP(A=a1, B=1)

        # assert str(b) == "BOOP(A=TupleFact('HI', TupleFact('HI', 2)), B=1.0)"
        print(b)

        print(str(TF(b, 1)))


@njit(cache=True)
def _b_dict_insert_tup_fact():
    d = Dict.empty(TupleFact, i8)
    for i in range(N):
        hash(TF(str(i),i))
        # d[TupleFact(str(i),i)] = i
    return d

@pytest.mark.benchmark(group="tuple_fact")
def test_b_dict_insert_tup_fact(benchmark):
    _b_dict_insert_tup_fact()
    benchmark.pedantic(_b_dict_insert_tup_fact,warmup_rounds=1,rounds=20)

def _b_dict_insert_tup():
    d = {}
    for i in range(N):
        hash((str(i),i))
        # d[(str(i),i)] = i
    return d

@pytest.mark.benchmark(group="tuple_fact")
def test_b_dict_insert_py_tup(benchmark):
    _b_dict_insert_tup()
    benchmark.pedantic(_b_dict_insert_tup,warmup_rounds=1,rounds=20)
    

@njit(cache=True)
def _b_make_tup_fact():
    d = Dict.empty(TupleFact, i8)
    for i in range(N):
        TF(str(i),i)
        # d[TupleFact(str(i),i)] = i
    return d

@pytest.mark.benchmark(group="tuple_fact")
def test_b_make_TupleFact(benchmark):
    _b_make_tup_fact()
    benchmark.pedantic(_b_make_tup_fact,warmup_rounds=1,rounds=20)    


@njit(cache=True)
def _b_make_py_tup():
    d = Dict.empty(TupleFact, i8)
    for i in range(N):
        (str(i),i)
        # d[TupleFact(str(i),i)] = i
    return d

@pytest.mark.benchmark(group="tuple_fact")
def test_b_make_py_tup(benchmark):
    _b_make_py_tup()
    benchmark.pedantic(_b_make_py_tup,warmup_rounds=1,rounds=20)    





if __name__ == "__main__":
    test_init()
    test_eq()
    test_hash()
    test_hash_obj_builtin_members()
    test_eq_obj_builtin_members()
    test_str()
    # p = init_prim()
    # print(type(p), p)
    # print(assert_cre_obj(2))
    # _b_make_tup_fact()
    # _b_dict_insert_tup_fact()

    
    # _b_make_py_tup()




import numpy as np
from numba.core import types
from numba import njit, float64, f8
from numba.experimental import structref
from numba.core.extending import overload_method, overload

@structref.register
class MyClassType(types.StructRef):
    pass

MyType = MyClassType([("x",f8[::1])])


class MyClass(structref.StructRefProxy):
    def __new__(cls, x):
        self = my_class_ctor(x)
        return self

    @property
    def x(self):
        return _x(self)

@njit(cache=True)
def _x(self):
    return self.x

@overload(MyClass)
def overload_MyClass(x):
    def impl(x):
        return my_class_ctor(x)
    return impl

structref.define_boxing(MyClassType, MyClass)

@njit(MyType(f8[::1]))
def my_class_ctor(x):
    st = structref.new(MyType)
    st.x = x
    return st

# This works...
@njit
def test():
    my_instance = MyClass(np.array([1.0, 2.0, 3.0]))
    print(my_instance.x) # prints [1. 2. 3.]
test()

# But this gets:
# TypeError: cannot convert native 
# numba.MyClassType(('x', array(float64, 1d, A)),) to Python object
my_instance = MyClass(np.array([1.0, 2.0, 3.0]))



import numpy as np
from numba import i8, i4, u8, u1, u2, u4, types, njit, generated_jit, literal_unroll
from numba.types import FunctionType, unicode_type, Tuple
from numba.extending import  overload, lower_getattr, overload_method
from cre.core import register_global_default, T_ID_UNDEFINED, T_ID_BOOL, T_ID_INT, T_ID_FLOAT, T_ID_STR, T_ID_TUPLE_FACT 
from cre.utils import (cast, _memcpy_structref, _obj_cast_codegen, ptr_t,
    _raw_ptr_from_struct_incref, _incref_ptr,
    CastFriendlyMixin, decode_idrec, _func_from_address, _incref_structref,
    _get_member_offset, _struct_get_data_ptr, _store, _store_safe,
    _sizeof_type, _load_ptr, encode_idrec, _decref_ptr, _incref_ptr,
    _decref_structref, check_issue_6993, incref_meminfo)
from cre.structref import define_structref
from numba.core.datamodel import default_manager, models
from numba.core import cgutils
from numba.experimental.structref import define_attributes, StructRefProxy, new, _Utils, define_boxing
from numba.experimental import structref
from numba.core import types, imputils, cgutils
from numba.core.datamodel import default_manager, models
from numba.core.extending import (
    models,
    intrinsic,
    lower_cast,
    lower_builtin,
    register_model,

    # infer_getattr,
    # lower_getattr_generic,
    # lower_setattr_generic,
    box,
    unbox,
    NativeValue,
    intrinsic,
    overload,
)

import operator


member_info_type = types.Tuple((u2,u2,u2))

# identity_members_info_type = {
#     "data" : types.UniTuple(member_info_type,1),
# }

# class IdentityMemberInfos(StructRefProxy):
#     @property
#     def data(self):
#         return identity_members_info_get_data(self)
#     # def __init__(self, fields):
#     #     super().__init__(fields)

# @structref.register
# class IdentityMemberInfosType(types.StructRef):
#     pass
#     # def preprocess_fields(self, fields):
#     #     if(isinstance(fields, int)):
#     #         return [('data',types.UniTuple(member_info_type, n))]

# def id_member_info_type(n):
#     # if(n <= 0): n =1
#     return IdentityMemberInfosType([('data',types.UniTuple(member_info_type, n))])


# BaseIdentityMemberInfosType = id_member_info_type(1)   

# define_attributes(IdentityMemberInfosType)
# define_boxing(IdentityMemberInfosType, IdentityMemberInfos)

# @njit(cache=True)
# def identity_members_info_get_data(self):
#     return self.data


# @overload(IdentityMemberInfos)
# @generated_jit(cache=True)
# def id_members_info_ctor(data):

#     # if(not isinstance(data, types.UniTuple)): return
#     n = data.count    

#     typ = id_member_info_type(n)   
#     def impl(data):
#         st = new(typ)
#         st.data = data
#         return _cast_structref(BaseIdentityMemberInfosType,st) 

#     return impl



# print(id_members_info_ctor(((0,0),(1,2))).data)




# chr_mbrs_info
from numba.cpython.hashing import _Py_hash_t
cre_obj_field_dict = {
    "idrec" : u8,
    # The number of members in the CREObj 
    "hash_val" : _Py_hash_t,
    "num_chr_mbrs": u4,
    "chr_mbrs_infos_offset" : u4,
    # The data offset of the "members" attribute (unpredictable because of layout alignment)
    # "chr_mbrs_infos" : BaseIdentityMemberInfosType,#types.UniTuple(member_info_type,1),
}

cre_obj_fields = [(k,v) for k,v in cre_obj_field_dict.items()]


class CREObjModel(models.StructRefModel):
    pass

def impl_cre_obj_upcast(context, builder, fromty, toty, val):
    return _obj_cast_codegen(context, builder, val, fromty, toty,incref=False)

class CREObjTypeClass(CastFriendlyMixin, types.StructRef):
    def __init__(self, fields,*args,**kwargs):
        if(isinstance(fields,dict)): fields = [(k,v) for k,v in fields.items()]
        types.StructRef.__init__(self,fields)
        # self.name = f'CREObjType'

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        # Allow numba type inferencer to upcast this type to CREObjType
        lower_cast(cls, CREObjType)(impl_cre_obj_upcast)
        

    def __str__(self):
        return f"CREObjType"

    def __repr__(self):
        return f"CREObjType"



    def hrepr():
        '''A human-readable repr for cre objects'''
        return f"{str(self)}(fields=({', '.join([f'{k}={str(v)}' for k,v in self._fields])}))" 


    # def __repr__(self):
    #     '''Make repr more readable than'''
    #     return f"{str(self)}(fields=({', '.join([f'{k}={str(v)}' for k,v in self._fields])}))" 


# @lower_cast(types.MemInfoPointer, types.Integer)
# def cast_meminfo_to_ptr(context, builder, fromty, toty, val):
#     return builder.ptrtoint(val, cgutils.intp_t)


CREObjType = CREObjTypeClass(cre_obj_fields) 
register_global_default("CREObj", CREObjType)
default_manager.register(CREObjTypeClass, CREObjModel)
define_attributes(CREObjType) #Not sure why but define on type instead of type_class


meminfo_type = types.MemInfoPointer(types.voidptr)
@njit(u2(meminfo_type),cache=True)
def get_t_id(ptr):
    cre_obj = cast(ptr, CREObjType)
    t_id, _, _ = decode_idrec(cre_obj.idrec)
    return t_id


from cre.utils import PrintElapse
class CREObjProxy(StructRefProxy):
    @classmethod
    def _numba_box_(cls, ty, mi):
        instance = ty.__new__(cls)
        instance._type = ty
        instance._meminfo = mi
        instance.recover_type_safe()
        
        
        return instance

    # __numba_ctor = cre_obj_ctor
    # def __new__(cls, idrec):
    #     return cre_obj_ctor(idrec)
    def __str__(self):
        return f"<cre.CREObj at {self.get_ptr()}>"

    # def get_ptr(self):
    #     return get_var_ptr(self)

    def get_ptr(self):
        return get_cre_obj_ptr(self)

    def get_ptr_incref(self):
        return get_cre_obj_ptr_incref(self)

    def asa(self, typ):
        if(typ._proxy_class is self._type): return self
        incref_meminfo(self._meminfo)
        proxy_typ = typ._proxy_class
        instance = super(StructRefProxy, proxy_typ).__new__(proxy_typ)
        instance._type = typ
        instance._meminfo = self._meminfo
        return instance
        # return asa(self,typ)

    def recover_type(self,context=None):
        ''' Try to recover the true type that the object was instantiated as'''            
        from cre.context import cre_context
        context = cre_context(context)# 2us

        t_id = get_t_id(self._meminfo)# .5us

        # print("<<", t_id, getattr(self._type,'t_id',0))
        # print(type(self),self._type,)
        if(getattr(self._type,'t_id',0) == t_id):# .5us
            return self

        # print("RECOVER",getattr(self._type,'t_id',0), t_id)
        if(t_id != T_ID_TUPLE_FACT):
            # The type associated with the object's t_id in 'context'
            t_id_type = context.get_type(t_id=t_id) # 3.9us
            if(self.__class__ is not t_id_type and hasattr(t_id_type,"_proxy_class")):
                self._type = t_id_type
                self.__class__ = t_id_type._proxy_class
        else:
            from cre.tuple_fact import define_tuple_fact
            mbr_t_ids = cre_obj_get_member_t_ids(self)
            tf_mbr_types = [context.get_type(t_id=t_id) for t_id in mbr_t_ids]
            tf_type,tf_proxy_type = define_tuple_fact(tf_mbr_types,context,return_proxy=True)
            self.__class__ = tf_proxy_type
        return self

    def recover_type_safe(self,context=None):
        try:
            return self.recover_type(context)
        except (ValueError, AttributeError) as e:
            print("RECOVER FAILED", e)
            return self

    # def __hash__(self):
    #     from cre.dynamic_exec import cre_obj_hash        
    #     return cre_obj_hash(self)

    def __eq__(self, other):
        from cre.dynamic_exec import cre_obj_eq
        if(not isinstance(other, CREObjProxy)): return False
        return cre_obj_eq(self, other)


CREObjType._proxy_class = CREObjProxy
define_boxing(CREObjTypeClass, CREObjProxy)


@njit(i8(CREObjType,), cache=True)
def get_cre_obj_ptr(self):
    return cast(self, i8)

@njit(i8(CREObjType,), cache=True)
def get_cre_obj_ptr_incref(self):
    return _raw_ptr_from_struct_incref(self)

@njit(u8(CREObjType,), cache=True)
def get_cre_obj_idrec(self):
    return self.idrec

PRIMITIVE_MBR_ID = 0
OBJECT_MBR_ID = 1
LIST_MBR_ID = 2
DICT_MBR_ID = 3

def resolve_member_id(x):
    if(isinstance(x, types.StructRef)):
        return OBJECT_MBR_ID
    elif(isinstance(x, types.ListType)):
        return LIST_MBR_ID
    elif(isinstance(x, types.DictType)):
        return DICT_MBR_ID
    else:
        return PRIMITIVE_MBR_ID

@intrinsic
def _get_chr_mbrs_infos_from_attrs(typingctx, structref, attrs_lit):
    '''get the base address of the struct pointed to by structref 'inst' '''
    # assert isinstance(attrs_lit, types.Literal)
    from cre.context import CREContext
    context = CREContext.get_default_context()

    # st_type = st_type_ref.instance_type
    # print("attrs_lit", attrs_lit)
    if(len(attrs_lit.types) >= 0 and
     not isinstance(attrs_lit.types[0],types.Literal)): return
    
    # print(attrs_lit.types)
    attrs = [x.literal_value for x in attrs_lit.types]
    # print(attrs)
    # print(ind)
    mbr_types = [v for k,v in structref._fields if k in attrs]
    t_ids = [context.get_t_id(_type=x) for x in mbr_types]
    m_ids = [resolve_member_id(x) for x in mbr_types]

    count = len(mbr_types)
    member_infos_out_type = types.UniTuple(member_info_type, count)
    # print(members_type.__dict__)


    def codegen(context, builder, sig, args):
        [st,_] = args
        utils = _Utils(context, builder, structref)

        baseptr = utils.get_data_pointer(st)
        baseptr_val = builder.ptrtoint(baseptr, cgutils.intp_t)
        dataval = utils.get_data_struct(st)

        member_infos = []
        # i = 0
        for i, (attr, typ) in enumerate(zip(attrs, mbr_types)):
            # print("<<", attr, base_fact_field_dict)
            # if(attr not in base_fact_field_dict):

            index_of_member = dataval._datamodel.get_field_position(attr)
            member_ptr = builder.gep(baseptr, [cgutils.int32_t(0), cgutils.int32_t(index_of_member)], inbounds=False)
            member_ptr = builder.ptrtoint(member_ptr, cgutils.intp_t)
            offset = builder.trunc(builder.sub(member_ptr, baseptr_val), cgutils.ir.IntType(16))
            t_id = context.get_constant(u2, t_ids[i])
            m_id = context.get_constant(u2, m_ids[i])
            member_infos.append(context.make_tuple(builder, member_info_type, (t_id, m_id, offset)))
            # i += 1

        ret = context.make_tuple(builder,member_infos_out_type, member_infos)
        return ret

    sig = member_infos_out_type(structref, attrs_lit)
    return sig, codegen


@njit
def set_chr_mbrs(st, chr_mbr_attrs):
    st.chr_mbrs_infos_offset = _get_member_offset(st,'chr_mbrs_infos')
    st.num_chr_mbrs =  len(chr_mbr_attrs)
    if(len(chr_mbr_attrs) > 0):
        chr_mbrs_infos = _get_chr_mbrs_infos_from_attrs(st, chr_mbr_attrs)
        st.chr_mbrs_infos = chr_mbrs_infos
    else:
        st.chr_mbrs_infos = ()
    st.hash_val = 0 
        
@njit(Tuple((u2,u2,i8))(CREObjType, i8),cache=True)
def cre_obj_get_item_t_id_ptr(x, index):
    data_ptr = _struct_get_data_ptr(x)
    member_info_ptr = data_ptr + x.chr_mbrs_infos_offset + index*_sizeof_type(member_info_type)
    t_id, m_id, member_offset = _load_ptr(member_info_type, member_info_ptr)
    return t_id, m_id, data_ptr + member_offset

@generated_jit(cache=True,nopython=True)
@overload_method(CREObjTypeClass,'get_item')
def cre_obj_get_item(obj, item_type, index):
    # print("obj", obj)
    # if(obj is not CREObjType): return
    def impl(obj, item_type, index):
        _, _, item_ptr = cre_obj_get_item_t_id_ptr(obj, index)
        out = _load_ptr(item_type, item_ptr)
        return out
    return impl

@generated_jit(cache=True,nopython=True)
def cre_obj_set_item(obj, index, val):
    item_type = val
    if(isinstance(item_type,types.Literal)):
        item_type = item_type.literal_type

    from cre.context import cre_context
    context = cre_context()

    m_id = resolve_member_id(item_type)
    t_id = context.get_t_id(_type=item_type)

    from numba.core.datamodel import default_manager, StructModel
    # if isinstance(default_manager[item_type], StructModel):
    # self._datamodel = self._context.data_model_manager[self._fe_type]
    def impl(obj, index, val):
        _, m_id, item_ptr = cre_obj_get_item_t_id_ptr(obj, index)
        # old_ptr = _load_ptr(i8, item_ptr)

        _store_safe(item_type, item_ptr, val)
        # _incref_structref(val)

        cre_obj_set_member_t_id_m_id(obj, index, (u2(t_id), u2(m_id)))

        # if(old_ptr != 0):
        #     _decref_ptr(old_ptr)
    # else:
    #     def impl(obj, index, val):
    #         _, m_id, item_ptr = cre_obj_get_item_t_id_ptr(obj, index)
    #         _store(item_type, item_ptr, val)        
    #         cre_obj_set_member_t_id_m_id(obj, index, (u2(t_id), u2(m_id)))
    return impl


# np_t_id_item_ptrs_type = np.dtype([('t_id', np.uint16),  ('m_id', np.uint16), ('ptr', np.int64)])
# t_id_item_ptrs = numba.from_dtype(np_t_id_item_ptrs_type)


# -------------------------------------------------------------------
# : _iter_mbr_infos()


mbr_info_tup_type = types.Tuple((u2, u2, i8))

class MbrInfoIterator(types.SimpleIteratorType):
    def __init__(self):
        name = f"MbrInfoIterator"
        super().__init__(name, mbr_info_tup_type)

MbrInfoIteratorType = MbrInfoIterator()

@register_model(MbrInfoIterator)
class MbrInfoIteratorModel(models.StructModel):
    def __init__(self, dmm, fe_type):
        members = [
            ('data_ptr', i8),
            ('member_info_ptr', types.EphemeralPointer(i8)),
            ('nitems', i4),
            ('index', types.EphemeralPointer(i4))
        ]
        super().__init__(dmm, fe_type, members)

from numba.core.imputils import lower_builtin, iternext_impl, RefType, impl_ret_borrowed

@intrinsic
def _iter_mbr_infos(typingctx, cre_obj):
    def codegen(context, builder, sig, args):
        (cre_obj,) = args
        # Make proxies for cre_obj and _iter
        cre_obj = cgutils.create_struct_proxy(CREObjType)(context, builder, value=args[0])
        _iter = cgutils.create_struct_proxy(MbrInfoIteratorType)(context, builder)

        # Set data_ptr
        raw_data_ptr = context.nrt.meminfo_data(builder, cre_obj.meminfo)
        _iter.data_ptr = builder.ptrtoint(raw_data_ptr, cgutils.intp_t)

        # Make struct proxy for underlying data
        valtype = CREObjType.get_data_type()
        model = context.data_model_manager[valtype]
        alloc_type = model.get_value_type()
        data_ptr = builder.bitcast(raw_data_ptr, alloc_type.as_pointer())
        data_struct = cgutils.create_struct_proxy(valtype)(
            context, builder, ref=data_ptr)

        # Extract and assign nitems
        nitems = builder.zext(getattr(data_struct, "num_chr_mbrs"), cgutils.int32_t)
        _iter.nitems = nitems

        # Extract chr_mbrs_infos_offset, assign member_info_ptr
        chr_mbrs_infos_offset = builder.zext(getattr(data_struct, "chr_mbrs_infos_offset"), cgutils.intp_t)
        data_ptr = builder.ptrtoint(data_ptr, cgutils.intp_t)
        member_info_ptr = builder.add(data_ptr, chr_mbrs_infos_offset)
        _iter.member_info_ptr = cgutils.alloca_once_value(builder, member_info_ptr)
        
        # Set index to 0.
        index = context.get_constant(types.int32, 0)
        _iter.index = cgutils.alloca_once_value(builder, index)

        return _iter._getvalue()
    sig = MbrInfoIteratorType(CREObjType)
    return sig, codegen


@lower_builtin('iternext', MbrInfoIterator)
@iternext_impl(RefType.BORROWED)
def iternext_listiter(context, builder, sig, args, result):
    # Define Constants
    llty = context.get_data_type(member_info_type)
    member_info_ptr_llty = llty.as_pointer()
    member_info_size = cgutils.sizeof(builder, member_info_ptr_llty)
    u2_ptr = context.get_data_type(u2).as_pointer()

    # Make iter proxy
    _iter = cgutils.create_struct_proxy(sig.args[0])(context, builder, value=args[0])

    # Load index and check inbounds.
    index = builder.load(_iter.index)
    is_valid = builder.icmp_signed('<', index, _iter.nitems)
    result.set_valid(is_valid)

    # If index is inbounds yeild and iterate 
    with builder.if_then(is_valid):
        member_info_ptr = builder.load(_iter.member_info_ptr)

        # Extract t_id
        t_id_ptr = builder.inttoptr(member_info_ptr, u2_ptr)
        t_id = builder.load(t_id_ptr)

        # Extract m_id
        m_id_ptr = builder.add(member_info_ptr, context.get_constant(i8, 2))
        m_id_ptr = builder.inttoptr(m_id_ptr, u2_ptr)
        m_id = builder.load(m_id_ptr)

        # Extract offset
        offset_ptr = builder.add(member_info_ptr, context.get_constant(i8, 4))
        offset_ptr = builder.inttoptr(offset_ptr, u2_ptr)
        offset = builder.load(offset_ptr)

        # DO: yield t_id, m_id, data_ptr + offset
        mbr_pointer = builder.add(_iter.data_ptr, builder.zext(offset, cgutils.intp_t))
        out = context.make_tuple(builder,mbr_info_tup_type,[t_id, m_id, mbr_pointer])
        result.yield_(out)

        # DO: _iter.index += 1
        index = builder.add(index, context.get_constant(i4, 1))
        builder.store(index, _iter.index)

        # DO:_iter.member_info_ptr += member_info_size1
        member_info_ptr = builder.load(_iter.member_info_ptr)
        member_info_ptr = builder.add(member_info_ptr, member_info_size)
        builder.store(member_info_ptr, _iter.member_info_ptr)
    

# has_not_fixed_6993 = check_issue_6993()

# @generated_jit(cache=False)
# def cre_obj_iter_t_id_item_ptrs(x):
#     def impl(x):
#         # x = _cast_structref(TupleFact,_x)
#         data_ptr = _struct_get_data_ptr(x)
#         member_info_ptr = data_ptr + x.chr_mbrs_infos_offset

#         # NOTE: Using as generator causes memory leak
#         for i in range(x.num_chr_mbrs):
#             t_id, m_id, member_offset = _load_ptr(member_info_type, member_info_ptr)
#             yield t_id, m_id, data_ptr + member_offset
#             member_info_ptr += _sizeof_type(member_info_type)
#             # _decref_structref(x)

#         if(has_not_fixed_6993):
#             _decref_structref(x)
            
#     return impl

t_id_m_id_type = types.Tuple((u2,u2))

@njit(cache=True)
def cre_obj_set_member_t_id_m_id(x, i, t_id_m_id):
    data_ptr = _struct_get_data_ptr(x)
    member_info_ptr = data_ptr + x.chr_mbrs_infos_offset + i * _sizeof_type(member_info_type)
    _store(t_id_m_id_type, member_info_ptr, t_id_m_id)



@njit(u2[::1](CREObjType),cache=True)
def cre_obj_get_member_t_ids(x):
    t_ids = np.empty((x.num_chr_mbrs,),dtype=np.uint16)
    for i, (t_id,_, _) in enumerate(_iter_mbr_infos(x)):
        t_ids[i] = t_id
    return t_ids


@generated_jit(cache=True,nopython=True)
@overload_method(CREObjType, "asa")
def asa(self, typ):
    def impl(self, typ):
        return cast(self, typ)
    return impl


@generated_jit(cache=True,nopython=True)
def copy_cre_obj(fact):
    fact_type = fact
    def impl(fact):
        new_fact = _memcpy_structref(fact)

        a,b = cast(fact, CREObjType), cast(new_fact, CREObjType)

        t_id, _, _ = decode_idrec(a.idrec)
        b.idrec = encode_idrec(t_id,0,u1(-1))
        b.hash_val = 0 
        for info_a in _iter_mbr_infos(a):
            t_id_a, m_id_a, data_ptr_a = info_a
            # t_id_b, m_id_b, data_ptr_b = info_b

            if(m_id_a != PRIMITIVE_MBR_ID):
                obj_ptr = _load_ptr(i8, data_ptr_a)
                _incref_ptr(obj_ptr)
            elif(t_id_a == T_ID_STR):
                s = _load_ptr(unicode_type, data_ptr_a)
                _incref_structref(s)
                #_store(unicode_type, data_ptr_b, s)

        # Weird extra refcounts... as much as 4 extra if call cre_obj_iter_t_id_item_ptrs
        #  on the new_fact
        _decref_structref(new_fact)

        return new_fact
    return impl


@njit(types.void(CREObjType), cache=True)
def cre_obj_clear_refs(fact):
    for info_a in _iter_mbr_infos(fact):
        t_id_a, m_id_a, data_ptr_a = info_a
        obj_ptr = _load_ptr(i8, data_ptr_a)
        if(m_id_a == OBJECT_MBR_ID):
            obj_ptr = _load_ptr(i8, data_ptr_a)
            _decref_ptr(obj_ptr)
            _store(i8, data_ptr_a, 0)
        # elif(t_id_a == T_ID_STR):
        #     s = _load_ptr(unicode_type, data_ptr_a)
        #     _incref_structref(s)
            
            



